#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright 2019 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

#############################################
#                WARNING                    #
#############################################
#
# This file is auto generated by the resource
#   module builder playbook.
#
# Do not edit this file manually.
#
# Changes to this file will be over written
#   by the resource module builder.
#
# Changes should be made in the model used to
#   generate this file or in the resource module
#   builder template.
#
#############################################
"""
The module file for exos_lldp_interfaces
"""

from __future__ import absolute_import, division, print_function
__metaclass__ = type

DOCUMENTATION = '''
---
module: exos_lldp_interfaces
version_added: '0.2.0'
short_description: Manage link layer discovery protocol (LLDP) attributes of interfaces on EXOS platforms.
description:
  - This module manages link layer discovery protocol (LLDP) attributes of interfaces on Extreme Networks EXOS platforms.
author: Jayalakshmi Viswanathan (@JayalakshmiV)
options:
  config:
    description: The list of link layer discovery protocol interface attribute configurations
    type: list
    elements: dict
    suboptions:
      name:
        description:
          - Name of the interface LLDP needs to be configured on.
        type: str
        required: True
      enabled:
        description:
          - This is a boolean value to control disabling of LLDP on the interface C(name)
        type: bool
  state:
    description:
      - The state the configuration should be left in.
    type: str
    choices:
      - merged
      - replaced
      - overridden
      - deleted
    default: merged
'''
EXAMPLES = """
# Using merged

# Before state:
# -------------
#
# path: /rest/restconf/data/openconfig-lldp:lldp/interfaces?depth=4
# method: GET
# data:
# {
#   "openconfig-lldp:interfaces": {
#     "interface": [
#       {
#         "config": {
#           "enabled": true,
#           "name": "1"
#         }
#       },
#       {
#         "config": {
#           "enabled": true,
#           "name": "2"
#         }
#       },
#       {
#         "config": {
#           "enabled": false,
#           "name": "3"
#         }
#       },
#       {
#         "config": {
#           "enabled": true,
#           "name": "4"
#         }
#       },
#       {
#         "config": {
#           "enabled": false,
#           "name": "5"
#         }
#       }
#     ]
#   }
# }

- name: Merge provided configuration with device configuration
  community.network.exos_lldp_interfaces:
    config:
      - name: '2'
        enabled: false
      - name: '5'
        enabled: true
    state: merged

# Module Execution Results:
# -------------------------
#
# "before":
#    - name: '1'
#      enabled: True
#    - name: '2'
#      enabled: True
#    - name: '3'
#      enabled: False
#    - name: '4'
#      enabled: True
#    - name: '5'
#      enabled: False
#
# "requests": [
#     {
#         "data": |
#         {
#           "openconfig-lldp:config": {
#             "enabled": false,
#             "name": "2"
#           }
#         }
#         "method": "PATCH",
#         "path": "/rest/restconf/data/openconfig-lldp:lldp/interfaces/interface=2/config"
#     },
#     {
#         "data": |
#         {
#           "openconfig-lldp:config": {
#             "enabled": true,
#             "name": "5"
#           }
#         }
#         "method": "PATCH",
#         "path": "/rest/restconf/data/openconfig-lldp:lldp/interfaces/interface=5/config"
#     }
# ]
#
# "after":
#    - name: '1'
#      enabled: True
#    - name: '2'
#      enabled: False
#    - name: '3'
#      enabled: False
#    - name: '4'
#      enabled: True
#    - name: '5'
#      enabled: True

# After state:
# -------------
#
# path: /rest/restconf/data/openconfig-lldp:lldp/interfaces?depth=4
# method: GET
# data:
# {
#   "openconfig-lldp:interfaces": {
#     "interface": [
#       {
#         "config": {
#           "enabled": true,
#           "name": "1"
#         }
#       },
#       {
#         "config": {
#           "enabled": false,
#           "name": "2"
#         }
#       },
#       {
#         "config": {
#           "enabled": false,
#           "name": "3"
#         }
#       },
#       {
#         "config": {
#           "enabled": true,
#           "name": "4"
#         }
#       },
#       {
#         "config": {
#           "enabled": true,
#           "name": "5"
#         }
#       }
#     ]
#   }
# }


# Using replaced

# Before state:
# -------------
#
# path: /rest/restconf/data/openconfig-lldp:lldp/interfaces?depth=4
# method: GET
# data:
# {
#   "openconfig-lldp:interfaces": {
#     "interface": [
#       {
#         "config": {
#           "enabled": true,
#           "name": "1"
#         }
#       },
#       {
#         "config": {
#           "enabled": true,
#           "name": "2"
#         }
#       },
#       {
#         "config": {
#           "enabled": false,
#           "name": "3"
#         }
#       },
#       {
#         "config": {
#           "enabled": true,
#           "name": "4"
#         }
#       },
#       {
#         "config": {
#           "enabled": false,
#           "name": "5"
#         }
#       }
#     ]
#   }
# }

- name: Replaces device configuration of listed lldp_interfaces with provided configuration
  community.network.exos_lldp_interfaces:
    config:
      - name: '1'
        enabled: false
      - name: '3'
        enabled: true
    state: merged

# Module Execution Results:
# -------------------------
#
# "before":
#    - name: '1'
#      enabled: True
#    - name: '2'
#      enabled: True
#    - name: '3'
#      enabled: False
#    - name: '4'
#      enabled: True
#    - name: '5'
#      enabled: False
#
# "requests": [
#     {
#         "data": |
#         {
#           "openconfig-lldp:config": {
#             "enabled": false,
#             "name": "1"
#           }
#         }
#         "method": "PATCH",
#         "path": "/rest/restconf/data/openconfig-lldp:lldp/interfaces/interface=1/config"
#     },
#     {
#         "data": |
#         {
#           "openconfig-lldp:config": {
#             "enabled": true,
#             "name": "3"
#           }
#         }
#         "method": "PATCH",
#         "path": "/rest/restconf/data/openconfig-lldp:lldp/interfaces/interface=3/config"
#     }
# ]
#
# "after":
#    - name: '1'
#      enabled: False
#    - name: '2'
#      enabled: True
#    - name: '3'
#      enabled: True
#    - name: '4'
#      enabled: True
#    - name: '5'
#      enabled: False

# After state:
# -------------
#
# path: /rest/restconf/data/openconfig-lldp:lldp/interfaces?depth=4
# method: GET
# data:
# {
#   "openconfig-lldp:interfaces": {
#     "interface": [
#       {
#         "config": {
#           "enabled": false,
#           "name": "1"
#         }
#       },
#       {
#         "config": {
#           "enabled": true,
#           "name": "2"
#         }
#       },
#       {
#         "config": {
#           "enabled": true,
#           "name": "3"
#         }
#       },
#       {
#         "config": {
#           "enabled": true,
#           "name": "4"
#         }
#       },
#       {
#         "config": {
#           "enabled": false,
#           "name": "5"
#         }
#       }
#     ]
#   }
# }


# Using deleted

# Before state:
# -------------
#
# path: /rest/restconf/data/openconfig-lldp:lldp/interfaces?depth=4
# method: GET
# data:
# {
#   "openconfig-lldp:interfaces": {
#     "interface": [
#       {
#         "config": {
#           "enabled": false,
#           "name": "1"
#         },
#       },
#       {
#         "config": {
#           "enabled": false,
#           "name": "2"
#         },
#       },
#       {
#         "config": {
#           "enabled": false,
#           "name": "3"
#         },
#       }
#     ]
#   }
# }

- name: Delete lldp interface configuration (this will not delete other lldp configuration)
  community.network.exos_lldp_interfaces:
    config:
      - name: '1'
      - name: '3'
    state: deleted

# Module Execution Results:
# -------------------------
#
# "before":
#    - name: '1'
#      enabled: False
#    - name: '2'
#      enabled: False
#    - name: '3'
#      enabled: False
#
# "requests": [
#     {
#         "data": |
#         {
#           "openconfig-lldp:config": {
#             "enabled": true,
#             "name": "1"
#           }
#         }
#         "method": "PATCH",
#         "path": "/rest/restconf/data/openconfig-lldp:lldp/interfaces/interface=1/config"
#     },
#     {
#         "data": |
#         {
#           "openconfig-lldp:config": {
#             "enabled": true,
#             "name": "3"
#           }
#         }
#         "method": "PATCH",
#         "path": "/rest/restconf/data/openconfig-lldp:lldp/interfaces/interface=3/config"
#     }
# ]
#
# "after":
#    - name: '1'
#      enabled: True
#    - name: '2'
#      enabled: False
#    - name: '3'
#      enabled: True
#
#  After state:
# -------------
# path: /rest/restconf/data/openconfig-lldp:lldp/interfaces?depth=4
# method: GET
# data:
# {
#   "openconfig-lldp:interfaces": {
#     "interface": [
#       {
#         "config": {
#           "enabled": true,
#           "name": "1"
#         },
#       },
#       {
#         "config": {
#           "enabled": false,
#           "name": "2"
#         },
#       },
#       {
#         "config": {
#           "enabled": true,
#           "name": "3"
#         },
#       }
#     ]
#   }
# }


# Using overridden

# Before state:
# -------------
#
# path: /rest/restconf/data/openconfig-lldp:lldp/interfaces?depth=4
# method: GET
# data:
# {
#   "openconfig-lldp:interfaces": {
#     "interface": [
#       {
#         "config": {
#           "enabled": true,
#           "name": "1"
#         }
#       },
#       {
#         "config": {
#           "enabled": true,
#           "name": "2"
#         }
#       },
#       {
#         "config": {
#           "enabled": false,
#           "name": "3"
#         }
#       },
#       {
#         "config": {
#           "enabled": true,
#           "name": "4"
#         }
#       },
#       {
#         "config": {
#           "enabled": false,
#           "name": "5"
#         }
#       }
#     ]
#   }
# }

- name: Override device configuration of all lldp_interfaces with provided configuration
  community.network.exos_lldp_interfaces:
    config:
      - name: '3'
        enabled: true
    state: overridden

# Module Execution Results:
# -------------------------
#
# "before":
#    - name: '1'
#      enabled: True
#    - name: '2'
#      enabled: True
#    - name: '3'
#      enabled: False
#    - name: '4'
#      enabled: True
#    - name: '5'
#      enabled: False
#
# "requests": [
#     {
#         "data": |
#         {
#           "openconfig-lldp:config": {
#             "enabled": true,
#             "name": "5"
#           }
#         }
#         "method": "PATCH",
#         "path": "/rest/restconf/data/openconfig-lldp:lldp/interfaces/interface=5/config"
#     },
#     {
#         "data": |
#         {
#           "openconfig-lldp:config": {
#             "enabled": true,
#             "name": "3"
#           }
#         }
#         "method": "PATCH",
#         "path": "/rest/restconf/data/openconfig-lldp:lldp/interfaces/interface=3/config"
#     }
# ]
#
# "after":
#    - name: '1'
#      enabled: True
#    - name: '2'
#      enabled: True
#    - name: '3'
#      enabled: True
#    - name: '4'
#      enabled: True
#    - name: '5'
#      enabled: True

# After state:
# -------------
#
# path: /rest/restconf/data/openconfig-lldp:lldp/interfaces?depth=4
# method: GET
# data:
# {
#   "openconfig-lldp:interfaces": {
#     "interface": [
#       {
#         "config": {
#           "enabled": true,
#           "name": "1"
#         }
#       },
#       {
#         "config": {
#           "enabled": true,
#           "name": "2"
#         }
#       },
#       {
#         "config": {
#           "enabled": true,
#           "name": "3"
#         }
#       },
#       {
#         "config": {
#           "enabled": true,
#           "name": "4"
#         }
#       },
#       {
#         "config": {
#           "enabled": true,
#           "name": "5"
#         }
#       }
#     ]
#   }
# }


"""
RETURN = """
before:
  description: The configuration prior to the model invocation.
  returned: always
  sample: >
    The configuration returned will always be in the same format
     of the parameters above.
  type: list
after:
  description: The resulting configuration model invocation.
  returned: when changed
  sample: >
    The configuration returned will always be in the same format
     of the parameters above.
  type: list
requests:
  description: The set of requests pushed to the remote device.
  returned: always
  type: list
  sample: [{"data": "...", "method": "...", "path": "..."}, {"data": "...", "method": "...", "path": "..."}, {"data": "...", "method": "...", "path": "..."}]
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.community.network.plugins.module_utils.network.exos.argspec.lldp_interfaces.lldp_interfaces import Lldp_interfacesArgs
from ansible_collections.community.network.plugins.module_utils.network.exos.config.lldp_interfaces.lldp_interfaces import Lldp_interfaces


def main():
    """
    Main entry point for module execution

    :returns: the result form module invocation
    """
    required_if = [('state', 'merged', ('config', )),
                   ('state', 'replaced', ('config', ))]
    module = AnsibleModule(argument_spec=Lldp_interfacesArgs.argument_spec,
                           required_if=required_if,
                           supports_check_mode=True)

    result = Lldp_interfaces(module).execute_module()
    module.exit_json(**result)


if __name__ == '__main__':
    main()



#!/usr/bin/python
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import, division, print_function

__metaclass__ = type


DOCUMENTATION = """
module: nxos_igmp_snooping
extends_documentation_fragment:
- cisco.nxos.nxos
short_description: Manages IGMP snooping global configuration.
description:
- Manages IGMP snooping global configuration.
version_added: 1.0.0
author:
- Jason Edelman (@jedelman8)
- Gabriele Gerbino (@GGabriele)
notes:
- Tested against NXOSv 7.3.(0)D1(1) on VIRL
- When C(state=default), params will be reset to a default state.
- C(group_timeout) also accepts I(never) as an input.
options:
  snooping:
    description:
    - Enables/disables IGMP snooping on the switch.
    type: bool
  group_timeout:
    description:
    - Group membership timeout value for all VLANs on the device. Accepted values
      are integer in range 1-10080, I(never) and I(default).
    type: str
  link_local_grp_supp:
    description:
    - Global link-local groups suppression.
    type: bool
  report_supp:
    description:
    - Global IGMPv1/IGMPv2 Report Suppression.
    type: bool
  v3_report_supp:
    description:
    - Global IGMPv3 Report Suppression and Proxy Reporting.
    type: bool
  state:
    description:
    - Manage the state of the resource.
    default: present
    choices:
    - present
    - default
    type: str
"""

EXAMPLES = """
# ensure igmp snooping params supported in this module are in there default state
- cisco.nxos.nxos_igmp_snooping:
    state: default

# ensure following igmp snooping params are in the desired state
- cisco.nxos.nxos_igmp_snooping:
    group_timeout: never
    snooping: true
    link_local_grp_supp: false
    optimize_mcast_flood: false
    report_supp: true
    v3_report_supp: true
"""

RETURN = """
commands:
    description: command sent to the device
    returned: always
    type: list
    sample: ["ip igmp snooping link-local-groups-suppression",
             "ip igmp snooping group-timeout 50",
             "no ip igmp snooping report-suppression",
             "no ip igmp snooping v3-report-suppression",
             "no ip igmp snooping"]
"""


import re

from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.nxos import (
    load_config,
    run_commands,
)
from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.nxos import (
    nxos_argument_spec,
)
from ansible.module_utils.basic import AnsibleModule


def execute_show_command(command, module, output="text"):
    command = {"command": command, "output": output}

    return run_commands(module, [command])


def flatten_list(command_lists):
    flat_command_list = []
    for command in command_lists:
        if isinstance(command, list):
            flat_command_list.extend(command)
        else:
            flat_command_list.append(command)
    return flat_command_list


def get_group_timeout(config):
    match = re.search(r"  Group timeout configured: (\S+)", config, re.M)
    if match:
        value = match.group(1)
    else:
        value = ""
    return value


def get_igmp_snooping(module):
    command = "show ip igmp snooping"
    existing = {}

    try:
        body = execute_show_command(command, module, output="json")[0]
    except IndexError:
        body = []

    if body:
        snooping = str(body.get("enabled")).lower()
        if snooping == "true" or snooping == "enabled":
            existing["snooping"] = True
        else:
            existing["snooping"] = False

        report_supp = str(body.get("grepsup")).lower()
        if report_supp == "true" or report_supp == "enabled":
            existing["report_supp"] = True
        else:
            existing["report_supp"] = False

        link_local_grp_supp = str(body.get("glinklocalgrpsup")).lower()
        if link_local_grp_supp == "true" or link_local_grp_supp == "enabled":
            existing["link_local_grp_supp"] = True
        else:
            existing["link_local_grp_supp"] = False

        v3_report_supp = str(body.get("gv3repsup")).lower()
        if v3_report_supp == "true" or v3_report_supp == "enabled":
            existing["v3_report_supp"] = True
        else:
            existing["v3_report_supp"] = False

    command = "show ip igmp snooping"
    body = execute_show_command(command, module)[0]
    if body:
        existing["group_timeout"] = get_group_timeout(body)

    return existing


def config_igmp_snooping(delta, existing, default=False):
    CMDS = {
        "snooping": "ip igmp snooping",
        "group_timeout": "ip igmp snooping group-timeout {}",
        "link_local_grp_supp": "ip igmp snooping link-local-groups-suppression",
        "v3_report_supp": "ip igmp snooping v3-report-suppression",
        "report_supp": "ip igmp snooping report-suppression",
    }

    commands = []
    command = None
    gt_command = None
    for key, value in delta.items():
        if value:
            if default and key == "group_timeout":
                if existing.get(key):
                    gt_command = "no " + CMDS.get(key).format(
                        existing.get(key)
                    )
            elif value == "default" and key == "group_timeout":
                if existing.get(key):
                    command = "no " + CMDS.get(key).format(existing.get(key))
            else:
                command = CMDS.get(key).format(value)
        else:
            command = "no " + CMDS.get(key).format(value)

        if command:
            commands.append(command)
        command = None

    if gt_command:
        # ensure that group-timeout command is configured last
        commands.append(gt_command)
    return commands


def get_igmp_snooping_defaults():
    group_timeout = "dummy"
    report_supp = True
    link_local_grp_supp = True
    v3_report_supp = False
    snooping = True

    args = dict(
        snooping=snooping,
        link_local_grp_supp=link_local_grp_supp,
        report_supp=report_supp,
        v3_report_supp=v3_report_supp,
        group_timeout=group_timeout,
    )

    default = dict(
        (param, value) for (param, value) in args.items() if value is not None
    )

    return default


def igmp_snooping_gt_dependency(command, existing, module):
    # group-timeout will fail if igmp snooping is disabled
    gt = [i for i in command if i.startswith("ip igmp snooping group-timeout")]
    if gt:
        if "no ip igmp snooping" in command or (
            existing["snooping"] is False and "ip igmp snooping" not in command
        ):
            msg = "group-timeout cannot be enabled or changed when ip igmp snooping is disabled"
            module.fail_json(msg=msg)
        else:
            # ensure that group-timeout command is configured last
            command.remove(gt[0])
            command.append(gt[0])


def main():
    argument_spec = dict(
        snooping=dict(required=False, type="bool"),
        group_timeout=dict(required=False, type="str"),
        link_local_grp_supp=dict(required=False, type="bool"),
        report_supp=dict(required=False, type="bool"),
        v3_report_supp=dict(required=False, type="bool"),
        state=dict(choices=["present", "default"], default="present"),
    )

    argument_spec.update(nxos_argument_spec)

    module = AnsibleModule(
        argument_spec=argument_spec, supports_check_mode=True
    )

    warnings = list()
    results = {"changed": False, "commands": [], "warnings": warnings}

    snooping = module.params["snooping"]
    link_local_grp_supp = module.params["link_local_grp_supp"]
    report_supp = module.params["report_supp"]
    v3_report_supp = module.params["v3_report_supp"]
    group_timeout = module.params["group_timeout"]
    state = module.params["state"]

    args = dict(
        snooping=snooping,
        link_local_grp_supp=link_local_grp_supp,
        report_supp=report_supp,
        v3_report_supp=v3_report_supp,
        group_timeout=group_timeout,
    )

    proposed = dict(
        (param, value) for (param, value) in args.items() if value is not None
    )

    existing = get_igmp_snooping(module)

    commands = []
    if state == "present":
        delta = dict(set(proposed.items()).difference(existing.items()))
        if delta:
            command = config_igmp_snooping(delta, existing)
            if command:
                if group_timeout:
                    igmp_snooping_gt_dependency(command, existing, module)
                commands.append(command)
    elif state == "default":
        proposed = get_igmp_snooping_defaults()
        delta = dict(set(proposed.items()).difference(existing.items()))
        if delta:
            command = config_igmp_snooping(delta, existing, default=True)
            if command:
                commands.append(command)

    cmds = flatten_list(commands)
    if cmds:
        results["changed"] = True
        if not module.check_mode:
            load_config(module, cmds)
        if "configure" in cmds:
            cmds.pop(0)
        results["commands"] = cmds

    module.exit_json(**results)


if __name__ == "__main__":
    main()



# (c) 2014, Brian Coca, Josh Drake, et al
# (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

DOCUMENTATION = '''
    author: Unknown (!UNKNOWN)
    cache: memcached
    short_description: Use memcached DB for cache
    description:
        - This cache uses JSON formatted, per host records saved in memcached.
    requirements:
      - memcache (python lib)
    options:
      _uri:
        description:
          - List of connection information for the memcached DBs
        default: ['127.0.0.1:11211']
        type: list
        env:
          - name: ANSIBLE_CACHE_PLUGIN_CONNECTION
        ini:
          - key: fact_caching_connection
            section: defaults
      _prefix:
        description: User defined prefix to use when creating the DB entries
        default: ansible_facts
        env:
          - name: ANSIBLE_CACHE_PLUGIN_PREFIX
        ini:
          - key: fact_caching_prefix
            section: defaults
      _timeout:
        default: 86400
        description: Expiration timeout in seconds for the cache plugin data. Set to 0 to never expire
        env:
          - name: ANSIBLE_CACHE_PLUGIN_TIMEOUT
        ini:
          - key: fact_caching_timeout
            section: defaults
        type: integer
'''

import collections
import os
import time
from multiprocessing import Lock
from itertools import chain

from ansible import constants as C
from ansible.errors import AnsibleError
from ansible.module_utils.common._collections_compat import MutableSet
from ansible.plugins.cache import BaseCacheModule
from ansible.utils.display import Display

try:
    import memcache
    HAS_MEMCACHE = True
except ImportError:
    HAS_MEMCACHE = False

display = Display()


class ProxyClientPool(object):
    """
    Memcached connection pooling for thread/fork safety. Inspired by py-redis
    connection pool.

    Available connections are maintained in a deque and released in a FIFO manner.
    """

    def __init__(self, *args, **kwargs):
        self.max_connections = kwargs.pop('max_connections', 1024)
        self.connection_args = args
        self.connection_kwargs = kwargs
        self.reset()

    def reset(self):
        self.pid = os.getpid()
        self._num_connections = 0
        self._available_connections = collections.deque(maxlen=self.max_connections)
        self._locked_connections = set()
        self._lock = Lock()

    def _check_safe(self):
        if self.pid != os.getpid():
            with self._lock:
                if self.pid == os.getpid():
                    # bail out - another thread already acquired the lock
                    return
                self.disconnect_all()
                self.reset()

    def get_connection(self):
        self._check_safe()
        try:
            connection = self._available_connections.popleft()
        except IndexError:
            connection = self.create_connection()
        self._locked_connections.add(connection)
        return connection

    def create_connection(self):
        if self._num_connections >= self.max_connections:
            raise RuntimeError("Too many memcached connections")
        self._num_connections += 1
        return memcache.Client(*self.connection_args, **self.connection_kwargs)

    def release_connection(self, connection):
        self._check_safe()
        self._locked_connections.remove(connection)
        self._available_connections.append(connection)

    def disconnect_all(self):
        for conn in chain(self._available_connections, self._locked_connections):
            conn.disconnect_all()

    def __getattr__(self, name):
        def wrapped(*args, **kwargs):
            return self._proxy_client(name, *args, **kwargs)
        return wrapped

    def _proxy_client(self, name, *args, **kwargs):
        conn = self.get_connection()

        try:
            return getattr(conn, name)(*args, **kwargs)
        finally:
            self.release_connection(conn)


class CacheModuleKeys(MutableSet):
    """
    A set subclass that keeps track of insertion time and persists
    the set in memcached.
    """
    PREFIX = 'ansible_cache_keys'

    def __init__(self, cache, *args, **kwargs):
        self._cache = cache
        self._keyset = dict(*args, **kwargs)

    def __contains__(self, key):
        return key in self._keyset

    def __iter__(self):
        return iter(self._keyset)

    def __len__(self):
        return len(self._keyset)

    def add(self, key):
        self._keyset[key] = time.time()
        self._cache.set(self.PREFIX, self._keyset)

    def discard(self, key):
        del self._keyset[key]
        self._cache.set(self.PREFIX, self._keyset)

    def remove_by_timerange(self, s_min, s_max):
        for k in self._keyset.keys():
            t = self._keyset[k]
            if s_min < t < s_max:
                del self._keyset[k]
        self._cache.set(self.PREFIX, self._keyset)


class CacheModule(BaseCacheModule):

    def __init__(self, *args, **kwargs):
        connection = ['127.0.0.1:11211']

        try:
            super(CacheModule, self).__init__(*args, **kwargs)
            if self.get_option('_uri'):
                connection = self.get_option('_uri')
            self._timeout = self.get_option('_timeout')
            self._prefix = self.get_option('_prefix')
        except KeyError:
            display.deprecated('Rather than importing CacheModules directly, '
                               'use ansible.plugins.loader.cache_loader',
                               version='2.0.0', collection_name='community.general')  # was Ansible 2.12
            if C.CACHE_PLUGIN_CONNECTION:
                connection = C.CACHE_PLUGIN_CONNECTION.split(',')
            self._timeout = C.CACHE_PLUGIN_TIMEOUT
            self._prefix = C.CACHE_PLUGIN_PREFIX

        if not HAS_MEMCACHE:
            raise AnsibleError("python-memcached is required for the memcached fact cache")

        self._cache = {}
        self._db = ProxyClientPool(connection, debug=0)
        self._keys = CacheModuleKeys(self._db, self._db.get(CacheModuleKeys.PREFIX) or [])

    def _make_key(self, key):
        return "{0}{1}".format(self._prefix, key)

    def _expire_keys(self):
        if self._timeout > 0:
            expiry_age = time.time() - self._timeout
            self._keys.remove_by_timerange(0, expiry_age)

    def get(self, key):
        if key not in self._cache:
            value = self._db.get(self._make_key(key))
            # guard against the key not being removed from the keyset;
            # this could happen in cases where the timeout value is changed
            # between invocations
            if value is None:
                self.delete(key)
                raise KeyError
            self._cache[key] = value

        return self._cache.get(key)

    def set(self, key, value):
        self._db.set(self._make_key(key), value, time=self._timeout, min_compress_len=1)
        self._cache[key] = value
        self._keys.add(key)

    def keys(self):
        self._expire_keys()
        return list(iter(self._keys))

    def contains(self, key):
        self._expire_keys()
        return key in self._keys

    def delete(self, key):
        del self._cache[key]
        self._db.delete(self._make_key(key))
        self._keys.discard(key)

    def flush(self):
        for key in self.keys():
            self.delete(key)

    def copy(self):
        return self._keys.copy()

    def __getstate__(self):
        return dict()

    def __setstate__(self, data):
        self.__init__()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2018, 2019 Kevin Breit (@kbreit) <kevin.breit@kevinbreit.net>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

ANSIBLE_METADATA = {
    'metadata_version': '1.1',
    'status': ['preview'],
    'supported_by': 'community'
}

DOCUMENTATION = r'''
---
module: meraki_alert
version_added: "2.1.0"
short_description: Manage alerts in the Meraki cloud
description:
- Allows for creation, management, and visibility into alert settings within Meraki.
options:
    state:
        description:
        - Create or modify an alert.
        choices: [ present, query ]
        default: present
        type: str
    net_name:
        description:
        - Name of a network.
        aliases: [ name, network ]
        type: str
    net_id:
        description:
        - ID number of a network.
        type: str
    default_destinations:
        description:
        - Properties for destinations when alert specific destinations aren't specified.
        type: dict
        suboptions:
            all_admins:
                description:
                - If true, all network admins will receive emails.
                type: bool
            snmp:
                description:
                - If true, then an SNMP trap will be sent if there is an SNMP trap server configured for this network.
                type: bool
            emails:
                description:
                - A list of emails that will recieve the alert(s).
                type: list
                elements: str
            http_server_ids:
                description:
                - A list of HTTP server IDs to send a Webhook to.
                type: list
                elements: str
    alerts:
        description:
        - Alert-specific configuration for each type.
        type: list
        elements: dict
        suboptions:
            alert_type:
                description:
                - The type of alert.
                type: str
            enabled:
                description:
                - A boolean depicting if the alert is turned on or off.
                type: bool
            filters:
                description:
                - A hash of specific configuration data for the alert. Only filters specific to the alert will be updated.
                - No validation checks occur against C(filters).
                type: raw
            alert_destinations:
                description:
                - A hash of destinations for this specific alert.
                type: dict
                suboptions:
                    all_admins:
                        description:
                        - If true, all network admins will receive emails.
                        type: bool
                    snmp:
                        description:
                        - If true, then an SNMP trap will be sent if there is an SNMP trap server configured for this network.
                        type: bool
                    emails:
                        description:
                        - A list of emails that will recieve the alert(s).
                        type: list
                        elements: str
                    http_server_ids:
                        description:
                        - A list of HTTP server IDs to send a Webhook to.
                        type: list
                        elements: str

author:
    - Kevin Breit (@kbreit)
extends_documentation_fragment: cisco.meraki.meraki
'''

EXAMPLES = r'''
- name: Update settings
  meraki_alert:
    auth_key: abc123
    org_name: YourOrg
    net_name: YourNet
    state: present
    default_destinations:
      emails:
      - 'youremail@yourcorp'
      - 'youremail2@yourcorp'
      all_admins: yes
      snmp: no
    alerts:
      - type: "gatewayDown"
        enabled: yes
        filters:
          timeout: 60
        alert_destinations:
          emails:
          - 'youremail@yourcorp'
          - 'youremail2@yourcorp'
          all_admins: yes
          snmp: no
      - type: "usageAlert"
        enabled: yes
        filters:
          period: 1200
          threshold: 104857600
        alert_destinations:
          emails:
          - 'youremail@yourcorp'
          - 'youremail2@yourcorp'
          all_admins: yes
          snmp: no

- name: Query all settings
  meraki_alert:
    auth_key: abc123
    org_name: YourOrg
    net_name: YourNet
    state: query
  delegate_to: localhost
'''

RETURN = r'''
data:
    description: Information about the created or manipulated object.
    returned: info
    type: complex
    contains:
        default_destinations:
            description: Properties for destinations when alert specific destinations aren't specified.
            returned: success
            type: complex
            contains:
                all_admins:
                    description: If true, all network admins will receive emails.
                    type: bool
                    sample: true
                    returned: success
                snmp:
                    description: If true, then an SNMP trap will be sent if there is an SNMP trap server configured for this network.
                    type: bool
                    sample: true
                    returned: success
                emails:
                    description: A list of emails that will recieve the alert(s).
                    type: list
                    returned: success
                http_server_ids:
                    description: A list of HTTP server IDs to send a Webhook to.
                    type: list
                    returned: success
        alerts:
            description: Alert-specific configuration for each type.
            type: complex
            contains:
                type:
                    description: The type of alert.
                    type: str
                    returned: success
                enabled:
                    description: A boolean depicting if the alert is turned on or off.
                    type: bool
                    returned: success
                filters:
                    description:
                    - A hash of specific configuration data for the alert. Only filters specific to the alert will be updated.
                    - No validation checks occur against C(filters).
                    type: complex
                    returned: success
                alert_destinations:
                    description: A hash of destinations for this specific alert.
                    type: complex
                    contains:
                        all_admins:
                            description: If true, all network admins will receive emails.
                            type: bool
                            returned: success
                        snmp:
                            description: If true, then an SNMP trap will be sent if there is an SNMP trap server configured for this network.
                            type: bool
                            returned: success
                        emails:
                            description: A list of emails that will recieve the alert(s).
                            type: list
                            returned: success
                        http_server_ids:
                            description: A list of HTTP server IDs to send a Webhook to.
                            type: list
                            returned: success
'''

import copy
from ansible.module_utils.basic import AnsibleModule, json
from ansible_collections.cisco.meraki.plugins.module_utils.network.meraki.meraki import MerakiModule, meraki_argument_spec


def construct_payload(meraki, original):
    payload = copy.deepcopy(original)
    if meraki.params['default_destinations'] is not None:
        payload['defaultDestinations'].update(meraki.params['default_destinations'])
        payload['defaultDestinations']['allAdmins'] = meraki.params['default_destinations']['all_admins']
        del payload['defaultDestinations']['all_admins']
        del payload['defaultDestinations']['http_server_ids']
    if meraki.params['alerts'] is not None:
        for alert in meraki.params['alerts']:
            alert.update(meraki.convert_snake_to_camel(alert))
            del alert['alert_destinations']
        for alert_want in meraki.params['alerts']:
            for alert_have in payload['alerts']:
                if alert_want['alert_type'] == alert_have['type']:
                    alert_have.update(alert_want)
                    del alert_have['alert_type']
                    del alert_have['alertType']
    return payload


def main():

    # define the available arguments/parameters that a user can pass to
    # the module

    destinations_arg_spec = dict(all_admins=dict(type='bool'),
                                 snmp=dict(type='bool'),
                                 emails=dict(type='list', elements='str'),
                                 http_server_ids=dict(type='list', elements='str', default=[]),
                                 )

    alerts_arg_spec = dict(alert_type=dict(type='str'),
                           enabled=dict(type='bool'),
                           alert_destinations=dict(type='dict', default=None, options=destinations_arg_spec),
                           filters=dict(type='raw', default={}),
                           )

    argument_spec = meraki_argument_spec()
    argument_spec.update(
        net_id=dict(type='str'),
        net_name=dict(type='str', aliases=['name', 'network']),
        state=dict(type='str', choices=['present', 'query'], default='present'),
        default_destinations=dict(type='dict', default=None, options=destinations_arg_spec),
        alerts=dict(type='list', elements='dict', options=alerts_arg_spec),
    )

    # the AnsibleModule object will be our abstraction working with Ansible
    # this includes instantiation, a couple of common attr would be the
    # args/params passed to the execution, as well as if the module
    # supports check mode
    module = AnsibleModule(argument_spec=argument_spec,
                           supports_check_mode=True,
                           )

    meraki = MerakiModule(module, function='alert')
    module.params['follow_redirects'] = 'all'

    query_urls = {'alert': '/networks/{net_id}/alerts/settings'}
    update_urls = {'alert': '/networks/{net_id}/alerts/settings'}
    meraki.url_catalog['get_all'].update(query_urls)
    meraki.url_catalog['update'] = update_urls

    # manipulate or modify the state as needed (this is going to be the
    # part where your module will do what it needs to do)

    org_id = meraki.params['org_id']
    if org_id is None:
        org_id = meraki.get_org_id(meraki.params['org_name'])
    net_id = meraki.params['net_id']
    if net_id is None:
        nets = meraki.get_nets(org_id=org_id)
        net_id = meraki.get_net_id(net_name=meraki.params['net_name'], data=nets)

    if meraki.params['state'] == 'query':
        path = meraki.construct_path('get_all', net_id=net_id)
        response = meraki.request(path, method='GET')
        if meraki.status == 200:
            meraki.result['data'] = response
        meraki.exit_json(**meraki.result)
    elif meraki.params['state'] == 'present':
        path = meraki.construct_path('get_all', net_id=net_id)
        original = meraki.request(path, method='GET')
        payload = construct_payload(meraki, original)
        # meraki.fail_json(msg="Compare", original=original, payload=payload)
        # meraki.fail_json(msg=payload)
        if meraki.is_update_required(original, payload):
            if meraki.check_mode is True:
                meraki.generate_diff(original, payload)
                meraki.result['data'] = payload
                meraki.result['changed'] = True
                meraki.exit_json(**meraki.result)
            path = meraki.construct_path('update', net_id=net_id)
            response = meraki.request(path, method='PUT', payload=json.dumps(payload))
            if meraki.status == 200:
                meraki.generate_diff(original, payload)
                meraki.result['data'] = response
                meraki.result['changed'] = True
                meraki.exit_json(**meraki.result)
        else:
            meraki.result['data'] = original
            meraki.exit_json(**meraki.result)

    # in the event of a successful module execution, you will want to
    # simple AnsibleModule.exit_json(), passing the key/value results
    meraki.exit_json(**meraki.result)


if __name__ == '__main__':
    main()



from __future__ import absolute_import, division, print_function
__metaclass__ = type

from . tower_module import TowerModule
from ansible.module_utils.urls import Request, SSLValidationError, ConnectionError
from ansible.module_utils.six import PY2
from ansible.module_utils.six.moves.urllib.parse import urlencode
from ansible.module_utils.six.moves.urllib.error import HTTPError
from ansible.module_utils.six.moves.http_cookiejar import CookieJar
import re
from json import loads, dumps


class TowerAPIModule(TowerModule):
    # TODO: Move the collection version check into tower_module.py
    # This gets set by the make process so whatever is in here is irrelevant
    _COLLECTION_VERSION = "14.1.0"
    _COLLECTION_TYPE = "awx"
    # This maps the collections type (awx/tower) to the values returned by the API
    # Those values can be found in awx/api/generics.py line 204
    collection_to_version = {
        'awx': 'AWX',
        'tower': 'Red Hat Ansible Tower',
    }
    session = None
    cookie_jar = CookieJar()

    def __init__(self, argument_spec, direct_params=None, error_callback=None, warn_callback=None, **kwargs):
        kwargs['supports_check_mode'] = True

        super(TowerAPIModule, self).__init__(argument_spec=argument_spec, direct_params=direct_params,
                                             error_callback=error_callback, warn_callback=warn_callback, **kwargs)
        self.session = Request(cookies=CookieJar(), validate_certs=self.verify_ssl)

    @staticmethod
    def param_to_endpoint(name):
        exceptions = {
            'inventory': 'inventories',
            'target_team': 'teams',
            'workflow': 'workflow_job_templates'
        }
        return exceptions.get(name, '{0}s'.format(name))

    def head_endpoint(self, endpoint, *args, **kwargs):
        return self.make_request('HEAD', endpoint, **kwargs)

    def get_endpoint(self, endpoint, *args, **kwargs):
        return self.make_request('GET', endpoint, **kwargs)

    def patch_endpoint(self, endpoint, *args, **kwargs):
        # Handle check mode
        if self.check_mode:
            self.json_output['changed'] = True
            self.exit_json(**self.json_output)

        return self.make_request('PATCH', endpoint, **kwargs)

    def post_endpoint(self, endpoint, *args, **kwargs):
        # Handle check mode
        if self.check_mode:
            self.json_output['changed'] = True
            self.exit_json(**self.json_output)

        return self.make_request('POST', endpoint, **kwargs)

    def delete_endpoint(self, endpoint, *args, **kwargs):
        # Handle check mode
        if self.check_mode:
            self.json_output['changed'] = True
            self.exit_json(**self.json_output)

        return self.make_request('DELETE', endpoint, **kwargs)

    def get_all_endpoint(self, endpoint, *args, **kwargs):
        response = self.get_endpoint(endpoint, *args, **kwargs)
        if 'next' not in response['json']:
            raise RuntimeError('Expected list from API at {0}, got: {1}'.format(endpoint, response))
        next_page = response['json']['next']

        if response['json']['count'] > 10000:
            self.fail_json(msg='The number of items being queried for is higher than 10,000.')

        while next_page is not None:
            next_response = self.get_endpoint(next_page)
            response['json']['results'] = response['json']['results'] + next_response['json']['results']
            next_page = next_response['json']['next']
            response['json']['next'] = next_page
        return response

    def get_one(self, endpoint, *args, **kwargs):
        response = self.get_endpoint(endpoint, *args, **kwargs)
        if response['status_code'] != 200:
            fail_msg = "Got a {0} response when trying to get one from {1}".format(response['status_code'], endpoint)
            if 'detail' in response.get('json', {}):
                fail_msg += ', detail: {0}'.format(response['json']['detail'])
            self.fail_json(msg=fail_msg)

        if 'count' not in response['json'] or 'results' not in response['json']:
            self.fail_json(msg="The endpoint did not provide count and results")

        if response['json']['count'] == 0:
            return None
        elif response['json']['count'] > 1:
            self.fail_json(msg="An unexpected number of items was returned from the API ({0})".format(response['json']['count']))

        return response['json']['results'][0]

    def resolve_name_to_id(self, endpoint, name_or_id):
        # Try to resolve the object by name
        name_field = 'name'
        if endpoint == 'users':
            name_field = 'username'

        response = self.get_endpoint(endpoint, **{'data': {name_field: name_or_id}})
        if response['status_code'] == 400:
            self.fail_json(msg="Unable to try and resolve {0} for {1} : {2}".format(endpoint, name_or_id, response['json']['detail']))

        if response['json']['count'] == 1:
            return response['json']['results'][0]['id']
        elif response['json']['count'] == 0:
            try:
                int(name_or_id)
                # If we got 0 items by name, maybe they gave us an ID, let's try looking it up by ID
                response = self.head_endpoint("{0}/{1}".format(endpoint, name_or_id), **{'return_none_on_404': True})
                if response is not None:
                    return name_or_id
            except ValueError:
                # If we got a value error than we didn't have an integer so we can just pass and fall down to the fail
                pass

            self.fail_json(msg="The {0} {1} was not found on the Tower server".format(endpoint, name_or_id))
        else:
            self.fail_json(msg="Found too many names {0} at endpoint {1} try using an ID instead of a name".format(name_or_id, endpoint))

    def make_request(self, method, endpoint, *args, **kwargs):
        # In case someone is calling us directly; make sure we were given a method, let's not just assume a GET
        if not method:
            raise Exception("The HTTP method must be defined")

        # Make sure we start with /api/vX
        if not endpoint.startswith("/"):
            endpoint = "/{0}".format(endpoint)
        if not endpoint.startswith("/api/"):
            endpoint = "/api/v2{0}".format(endpoint)
        if not endpoint.endswith('/') and '?' not in endpoint:
            endpoint = "{0}/".format(endpoint)

        # Extract the headers, this will be used in a couple of places
        headers = kwargs.get('headers', {})

        # Authenticate to Tower (if we don't have a token and if not already done so)
        if not self.oauth_token and not self.authenticated:
            # This method will set a cookie in the cookie jar for us and also an oauth_token
            self.authenticate(**kwargs)
        if self.oauth_token:
            # If we have a oauth token, we just use a bearer header
            headers['Authorization'] = 'Bearer {0}'.format(self.oauth_token)

        # Update the URL path with the endpoint
        self.url = self.url._replace(path=endpoint)

        if method in ['POST', 'PUT', 'PATCH']:
            headers.setdefault('Content-Type', 'application/json')
            kwargs['headers'] = headers
        elif kwargs.get('data'):
            self.url = self.url._replace(query=urlencode(kwargs.get('data')))

        data = None  # Important, if content type is not JSON, this should not be dict type
        if headers.get('Content-Type', '') == 'application/json':
            data = dumps(kwargs.get('data', {}))

        try:
            response = self.session.open(method, self.url.geturl(), headers=headers, validate_certs=self.verify_ssl, follow_redirects=True, data=data)
        except(SSLValidationError) as ssl_err:
            self.fail_json(msg="Could not establish a secure connection to your host ({1}): {0}.".format(self.url.netloc, ssl_err))
        except(ConnectionError) as con_err:
            self.fail_json(msg="There was a network error of some kind trying to connect to your host ({1}): {0}.".format(self.url.netloc, con_err))
        except(HTTPError) as he:
            # Sanity check: Did the server send back some kind of internal error?
            if he.code >= 500:
                self.fail_json(msg='The host sent back a server error ({1}): {0}. Please check the logs and try again later'.format(self.url.path, he))
            # Sanity check: Did we fail to authenticate properly?  If so, fail out now; this is always a failure.
            elif he.code == 401:
                self.fail_json(msg='Invalid Tower authentication credentials for {0} (HTTP 401).'.format(self.url.path))
            # Sanity check: Did we get a forbidden response, which means that the user isn't allowed to do this? Report that.
            elif he.code == 403:
                self.fail_json(msg="You don't have permission to {1} to {0} (HTTP 403).".format(self.url.path, method))
            # Sanity check: Did we get a 404 response?
            # Requests with primary keys will return a 404 if there is no response, and we want to consistently trap these.
            elif he.code == 404:
                if kwargs.get('return_none_on_404', False):
                    return None
                self.fail_json(msg='The requested object could not be found at {0}.'.format(self.url.path))
            # Sanity check: Did we get a 405 response?
            # A 405 means we used a method that isn't allowed. Usually this is a bad request, but it requires special treatment because the
            # API sends it as a logic error in a few situations (e.g. trying to cancel a job that isn't running).
            elif he.code == 405:
                self.fail_json(msg="The Tower server says you can't make a request with the {0} method to this endpoing {1}".format(method, self.url.path))
            # Sanity check: Did we get some other kind of error?  If so, write an appropriate error message.
            elif he.code >= 400:
                # We are going to return a 400 so the module can decide what to do with it
                page_data = he.read()
                try:
                    return {'status_code': he.code, 'json': loads(page_data)}
                # JSONDecodeError only available on Python 3.5+
                except ValueError:
                    return {'status_code': he.code, 'text': page_data}
            elif he.code == 204 and method == 'DELETE':
                # A 204 is a normal response for a delete function
                pass
            else:
                self.fail_json(msg="Unexpected return code when calling {0}: {1}".format(self.url.geturl(), he))
        except(Exception) as e:
            self.fail_json(msg="There was an unknown error when trying to connect to {2}: {0} {1}".format(type(e).__name__, e, self.url.geturl()))
        finally:
            self.url = self.url._replace(query=None)

        if not self.version_checked:
            # In PY2 we get back an HTTPResponse object but PY2 is returning an addinfourl
            # First try to get the headers in PY3 format and then drop down to PY2.
            try:
                tower_type = response.getheader('X-API-Product-Name', None)
                tower_version = response.getheader('X-API-Product-Version', None)
            except Exception:
                tower_type = response.info().getheader('X-API-Product-Name', None)
                tower_version = response.info().getheader('X-API-Product-Version', None)

            if self._COLLECTION_TYPE not in self.collection_to_version or self.collection_to_version[self._COLLECTION_TYPE] != tower_type:
                self.warn("You are using the {0} version of this collection but connecting to {1}".format(
                    self._COLLECTION_TYPE, tower_type
                ))
            elif self._COLLECTION_VERSION != tower_version:
                self.warn("You are running collection version {0} but connecting to tower version {1}".format(
                    self._COLLECTION_VERSION, tower_version
                ))
            self.version_checked = True

        response_body = ''
        try:
            response_body = response.read()
        except(Exception) as e:
            self.fail_json(msg="Failed to read response body: {0}".format(e))

        response_json = {}
        if response_body and response_body != '':
            try:
                response_json = loads(response_body)
            except(Exception) as e:
                self.fail_json(msg="Failed to parse the response json: {0}".format(e))

        if PY2:
            status_code = response.getcode()
        else:
            status_code = response.status
        return {'status_code': status_code, 'json': response_json}

    def authenticate(self, **kwargs):
        if self.username and self.password:
            # Attempt to get a token from /api/v2/tokens/ by giving it our username/password combo
            # If we have a username and password, we need to get a session cookie
            login_data = {
                "description": "Ansible Tower Module Token",
                "application": None,
                "scope": "write",
            }
            # Post to the tokens endpoint with baisc auth to try and get a token
            api_token_url = (self.url._replace(path='/api/v2/tokens/')).geturl()

            try:
                response = self.session.open(
                    'POST', api_token_url,
                    validate_certs=self.verify_ssl, follow_redirects=True,
                    force_basic_auth=True, url_username=self.username, url_password=self.password,
                    data=dumps(login_data), headers={'Content-Type': 'application/json'}
                )
            except HTTPError as he:
                try:
                    resp = he.read()
                except Exception as e:
                    resp = 'unknown {0}'.format(e)
                self.fail_json(msg='Failed to get token: {0}'.format(he), response=resp)
            except(Exception) as e:
                # Sanity check: Did the server send back some kind of internal error?
                self.fail_json(msg='Failed to get token: {0}'.format(e))

            token_response = None
            try:
                token_response = response.read()
                response_json = loads(token_response)
                self.oauth_token_id = response_json['id']
                self.oauth_token = response_json['token']
            except(Exception) as e:
                self.fail_json(msg="Failed to extract token information from login response: {0}".format(e), **{'response': token_response})

        # If we have neither of these, then we can try un-authenticated access
        self.authenticated = True

    def delete_if_needed(self, existing_item, on_delete=None):
        # This will exit from the module on its own.
        # If the method successfully deletes an item and on_delete param is defined,
        #   the on_delete parameter will be called as a method pasing in this object and the json from the response
        # This will return one of two things:
        #   1. None if the existing_item is not defined (so no delete needs to happen)
        #   2. The response from Tower from calling the delete on the endpont. It's up to you to process the response and exit from the module
        # Note: common error codes from the Tower API can cause the module to fail
        if existing_item:
            # If we have an item, we can try to delete it
            try:
                item_url = existing_item['url']
                item_type = existing_item['type']
                item_id = existing_item['id']
            except KeyError as ke:
                self.fail_json(msg="Unable to process delete of item due to missing data {0}".format(ke))

            if 'name' in existing_item:
                item_name = existing_item['name']
            elif 'username' in existing_item:
                item_name = existing_item['username']
            elif 'identifier' in existing_item:
                item_name = existing_item['identifier']
            elif item_type == 'o_auth2_access_token':
                # An oauth2 token has no name, instead we will use its id for any of the messages
                item_name = existing_item['id']
            elif item_type == 'credential_input_source':
                # An credential_input_source has no name, instead we will use its id for any of the messages
                item_name = existing_item['id']
            else:
                self.fail_json(msg="Unable to process delete of {0} due to missing name".format(item_type))

            response = self.delete_endpoint(item_url)

            if response['status_code'] in [202, 204]:
                if on_delete:
                    on_delete(self, response['json'])
                self.json_output['changed'] = True
                self.json_output['id'] = item_id
                self.exit_json(**self.json_output)
            else:
                if 'json' in response and '__all__' in response['json']:
                    self.fail_json(msg="Unable to delete {0} {1}: {2}".format(item_type, item_name, response['json']['__all__'][0]))
                elif 'json' in response:
                    # This is from a project delete (if there is an active job against it)
                    if 'error' in response['json']:
                        self.fail_json(msg="Unable to delete {0} {1}: {2}".format(item_type, item_name, response['json']['error']))
                    else:
                        self.fail_json(msg="Unable to delete {0} {1}: {2}".format(item_type, item_name, response['json']))
                else:
                    self.fail_json(msg="Unable to delete {0} {1}: {2}".format(item_type, item_name, response['status_code']))
        else:
            self.exit_json(**self.json_output)

    def modify_associations(self, association_endpoint, new_association_list):
        # if we got None instead of [] we are not modifying the association_list
        if new_association_list is None:
            return

        # First get the existing associations
        response = self.get_all_endpoint(association_endpoint)
        existing_associated_ids = [association['id'] for association in response['json']['results']]

        # Disassociate anything that is in existing_associated_ids but not in new_association_list
        ids_to_remove = list(set(existing_associated_ids) - set(new_association_list))
        for an_id in ids_to_remove:
            response = self.post_endpoint(association_endpoint, **{'data': {'id': int(an_id), 'disassociate': True}})
            if response['status_code'] == 204:
                self.json_output['changed'] = True
            else:
                self.fail_json(msg="Failed to disassociate item {0}".format(response['json']['detail']))

        # Associate anything that is in new_association_list but not in `association`
        for an_id in list(set(new_association_list) - set(existing_associated_ids)):
            response = self.post_endpoint(association_endpoint, **{'data': {'id': int(an_id)}})
            if response['status_code'] == 204:
                self.json_output['changed'] = True
            else:
                self.fail_json(msg="Failed to associate item {0}".format(response['json']['detail']))

    def create_if_needed(self, existing_item, new_item, endpoint, on_create=None, item_type='unknown', associations=None):

        # This will exit from the module on its own
        # If the method successfully creates an item and on_create param is defined,
        #    the on_create parameter will be called as a method pasing in this object and the json from the response
        # This will return one of two things:
        #    1. None if the existing_item is already defined (so no create needs to happen)
        #    2. The response from Tower from calling the patch on the endpont. It's up to you to process the response and exit from the module
        # Note: common error codes from the Tower API can cause the module to fail

        if not endpoint:
            self.fail_json(msg="Unable to create new {0} due to missing endpoint".format(item_type))

        item_url = None
        if existing_item:
            try:
                item_url = existing_item['url']
            except KeyError as ke:
                self.fail_json(msg="Unable to process create of item due to missing data {0}".format(ke))
        else:
            # If we don't have an exisitng_item, we can try to create it

            # We have to rely on item_type being passed in since we don't have an existing item that declares its type
            # We will pull the item_name out from the new_item, if it exists
            for key in ('name', 'username', 'identifier', 'hostname'):
                if key in new_item:
                    item_name = new_item[key]
                    break
            else:
                item_name = 'unknown'

            response = self.post_endpoint(endpoint, **{'data': new_item})
            if response['status_code'] == 201:
                self.json_output['name'] = 'unknown'
                for key in ('name', 'username', 'identifier', 'hostname'):
                    if key in response['json']:
                        self.json_output['name'] = response['json'][key]
                self.json_output['id'] = response['json']['id']
                self.json_output['changed'] = True
                item_url = response['json']['url']
            else:
                if 'json' in response and '__all__' in response['json']:
                    self.fail_json(msg="Unable to create {0} {1}: {2}".format(item_type, item_name, response['json']['__all__'][0]))
                elif 'json' in response:
                    self.fail_json(msg="Unable to create {0} {1}: {2}".format(item_type, item_name, response['json']))
                else:
                    self.fail_json(msg="Unable to create {0} {1}: {2}".format(item_type, item_name, response['status_code']))

        # Process any associations with this item
        if associations is not None:
            for association_type in associations:
                sub_endpoint = '{0}{1}/'.format(item_url, association_type)
                self.modify_associations(sub_endpoint, associations[association_type])

        # If we have an on_create method and we actually changed something we can call on_create
        if on_create is not None and self.json_output['changed']:
            on_create(self, response['json'])
        else:
            self.exit_json(**self.json_output)

    def _encrypted_changed_warning(self, field, old, warning=False):
        if not warning:
            return
        self.warn(
            'The field {0} of {1} {2} has encrypted data and may inaccurately report task is changed.'.format(
                field, old.get('type', 'unknown'), old.get('id', 'unknown')
            ))

    @staticmethod
    def has_encrypted_values(obj):
        """Returns True if JSON-like python content in obj has $encrypted$
        anywhere in the data as a value
        """
        if isinstance(obj, dict):
            for val in obj.values():
                if TowerAPIModule.has_encrypted_values(val):
                    return True
        elif isinstance(obj, list):
            for val in obj:
                if TowerAPIModule.has_encrypted_values(val):
                    return True
        elif obj == TowerAPIModule.ENCRYPTED_STRING:
            return True
        return False

    def objects_could_be_different(self, old, new, field_set=None, warning=False):
        if field_set is None:
            field_set = set(fd for fd in new.keys() if fd not in ('modified', 'related', 'summary_fields'))
        for field in field_set:
            new_field = new.get(field, None)
            old_field = old.get(field, None)
            if old_field != new_field:
                return True  # Something doesn't match
            elif self.has_encrypted_values(new_field) or field not in new:
                # case of 'field not in new' - user password write-only field that API will not display
                self._encrypted_changed_warning(field, old, warning=warning)
                return True
        return False

    def update_if_needed(self, existing_item, new_item, on_update=None, associations=None):
        # This will exit from the module on its own
        # If the method successfully updates an item and on_update param is defined,
        #   the on_update parameter will be called as a method pasing in this object and the json from the response
        # This will return one of two things:
        #    1. None if the existing_item does not need to be updated
        #    2. The response from Tower from patching to the endpoint. It's up to you to process the response and exit from the module.
        # Note: common error codes from the Tower API can cause the module to fail
        response = None
        if existing_item:

            # If we have an item, we can see if it needs an update
            try:
                item_url = existing_item['url']
                item_type = existing_item['type']
                if item_type == 'user':
                    item_name = existing_item['username']
                elif item_type == 'workflow_job_template_node':
                    item_name = existing_item['identifier']
                elif item_type == 'credential_input_source':
                    item_name = existing_item['id']
                else:
                    item_name = existing_item['name']
                item_id = existing_item['id']
            except KeyError as ke:
                self.fail_json(msg="Unable to process update of item due to missing data {0}".format(ke))

            # Check to see if anything within the item requires the item to be updated
            needs_patch = self.objects_could_be_different(existing_item, new_item)

            # If we decided the item needs to be updated, update it
            self.json_output['id'] = item_id
            if needs_patch:
                response = self.patch_endpoint(item_url, **{'data': new_item})
                if response['status_code'] == 200:
                    # compare apples-to-apples, old API data to new API data
                    # but do so considering the fields given in parameters
                    self.json_output['changed'] = self.objects_could_be_different(
                        existing_item, response['json'], field_set=new_item.keys(), warning=True)
                elif 'json' in response and '__all__' in response['json']:
                    self.fail_json(msg=response['json']['__all__'])
                else:
                    self.fail_json(**{'msg': "Unable to update {0} {1}, see response".format(item_type, item_name), 'response': response})

        else:
            raise RuntimeError('update_if_needed called incorrectly without existing_item')

        # Process any associations with this item
        if associations is not None:
            for association_type, id_list in associations.items():
                endpoint = '{0}{1}/'.format(item_url, association_type)
                self.modify_associations(endpoint, id_list)

        # If we change something and have an on_change call it
        if on_update is not None and self.json_output['changed']:
            if response is None:
                last_data = existing_item
            else:
                last_data = response['json']
            on_update(self, last_data)
        else:
            self.exit_json(**self.json_output)

    def create_or_update_if_needed(self, existing_item, new_item, endpoint=None, item_type='unknown', on_create=None, on_update=None, associations=None):
        if existing_item:
            return self.update_if_needed(existing_item, new_item, on_update=on_update, associations=associations)
        else:
            return self.create_if_needed(existing_item, new_item, endpoint, on_create=on_create, item_type=item_type, associations=associations)

    def logout(self):
        if self.authenticated:
            # Attempt to delete our current token from /api/v2/tokens/
            # Post to the tokens endpoint with baisc auth to try and get a token
            api_token_url = (
                self.url._replace(
                    path='/api/v2/tokens/{0}/'.format(self.oauth_token_id),
                    query=None  # in error cases, fail_json exists before exception handling
                )
            ).geturl()

            try:
                self.session.open(
                    'DELETE',
                    api_token_url,
                    validate_certs=self.verify_ssl,
                    follow_redirects=True,
                    force_basic_auth=True,
                    url_username=self.username,
                    url_password=self.password
                )
                self.oauth_token_id = None
                self.authenticated = False
            except HTTPError as he:
                try:
                    resp = he.read()
                except Exception as e:
                    resp = 'unknown {0}'.format(e)
                self.warn('Failed to release tower token: {0}, response: {1}'.format(he, resp))
            except(Exception) as e:
                # Sanity check: Did the server send back some kind of internal error?
                self.warn('Failed to release tower token {0}: {1}'.format(self.oauth_token_id, e))

    def is_job_done(self, job_status):
        if job_status in ['new', 'pending', 'waiting', 'running']:
            return False
        else:
            return True



# (c) 2020 Red Hat Inc.
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

# Make coding more python3-ish

from __future__ import absolute_import, division, print_function

__metaclass__ = type

from textwrap import dedent
from ansible_collections.cisco.nxos.tests.unit.compat.mock import patch
from ansible_collections.cisco.nxos.tests.unit.modules.utils import (
    AnsibleFailJson,
)
from ansible_collections.cisco.nxos.plugins.modules import nxos_ospfv3

from .nxos_module import TestNxosModule, load_fixture, set_module_args

ignore_provider_arg = True


class TestNxosOspfv3Module(TestNxosModule):

    # Testing strategy
    # ------------------
    # (a) The unit tests cover `merged` and `replaced` for every attribute.
    #     Since `overridden` is essentially `replaced` but at a larger
    #     scale, these indirectly cover `overridden` as well.
    # (b) For linear attributes replaced is not valid and hence, those tests
    #     delete the attributes from the config subsection.
    # (c) The argspec for VRFs is same as the top-level spec and the config logic
    #     is re-used. Hence, those attributes are not explictly covered. However, a
    #     combination of VRF + top-level spec + AF is tested.

    module = nxos_ospfv3

    def setUp(self):
        super(TestNxosOspfv3Module, self).setUp()

        self.mock_get_resource_connection = patch(
            "ansible_collections.ansible.netcommon.plugins.module_utils.network.common.resource_module.get_resource_connection"
        )
        self.get_resource_connection = (
            self.mock_get_resource_connection.start()
        )

        self.mock_get_config = patch(
            "ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.facts.ospfv3.ospfv3.Ospfv3Facts.get_config"
        )
        self.get_config = self.mock_get_config.start()

    def tearDown(self):
        super(TestNxosOspfv3Module, self).tearDown()
        self.get_resource_connection.stop()
        self.get_config.stop()

    def test_nxos_ospfv3_af_areas_filter_list_merged(self):
        # test merged for config->processes->af->areas->filter_list
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              address-family ipv6 unicast
                area 1.1.1.1 default-cost 100
                area 1.1.1.1 filter-list route-map test-11 in
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            address_family=dict(
                                afi="ipv6",
                                safi="unicast",
                                areas=[
                                    dict(
                                        area_id="1.1.1.1",
                                        filter_list=[
                                            dict(
                                                route_map="test-1",
                                                direction="in",
                                            ),
                                            dict(
                                                route_map="test-2",
                                                direction="out",
                                            ),
                                        ],
                                    ),
                                    dict(
                                        area_id="1.1.1.2",
                                        filter_list=[
                                            dict(
                                                route_map="test-3",
                                                direction="in",
                                            ),
                                            dict(
                                                route_map="test-4",
                                                direction="out",
                                            ),
                                        ],
                                    ),
                                ],
                            ),
                        )
                    ]
                ),
                state="merged",
            ),
            ignore_provider_arg,
        )
        commands = [
            "router ospfv3 100",
            "address-family ipv6 unicast",
            "area 1.1.1.1 filter-list route-map test-1 in",
            "area 1.1.1.1 filter-list route-map test-2 out",
            "area 1.1.1.2 filter-list route-map test-3 in",
            "area 1.1.1.2 filter-list route-map test-4 out",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_af_areas_filter_list_replaced(self):
        # test replaced for config->processes->af->areas->filter_list
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              address-family ipv6 unicast
                area 1.1.1.4 filter-list route-map test-11 out
                area 1.1.1.4 filter-list route-map test-12 in
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            address_family=dict(
                                afi="ipv6",
                                safi="unicast",
                                areas=[
                                    dict(
                                        area_id="1.1.1.1",
                                        filter_list=[
                                            dict(
                                                route_map="test-1",
                                                direction="in",
                                            ),
                                            dict(
                                                route_map="test-2",
                                                direction="out",
                                            ),
                                        ],
                                    ),
                                    dict(
                                        area_id="1.1.1.2",
                                        filter_list=[
                                            dict(
                                                route_map="test-3",
                                                direction="in",
                                            ),
                                            dict(
                                                route_map="test-4",
                                                direction="out",
                                            ),
                                        ],
                                    ),
                                ],
                            ),
                        )
                    ]
                ),
                state="replaced",
            ),
            ignore_provider_arg,
        )
        commands = [
            "router ospfv3 100",
            "address-family ipv6 unicast",
            "no area 1.1.1.4 filter-list route-map test-11 out",
            "no area 1.1.1.4 filter-list route-map test-12 in",
            "area 1.1.1.1 filter-list route-map test-1 in",
            "area 1.1.1.1 filter-list route-map test-2 out",
            "area 1.1.1.2 filter-list route-map test-3 in",
            "area 1.1.1.2 filter-list route-map test-4 out",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_af_areas_ranges_merged(self):
        # test merged for config->processes->af->areas->rang
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              address-family ipv6 unicast
                area 1.1.1.1 range 2001:db2::/32
                area 1.1.1.1 range 2001:db3::/32 cost 10
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            address_family=dict(
                                afi="ipv6",
                                safi="unicast",
                                areas=[
                                    dict(
                                        area_id="1.1.1.1",
                                        ranges=[
                                            dict(
                                                prefix="2001:db3::/32",
                                                cost="20",
                                            )
                                        ],
                                    ),
                                    dict(
                                        area_id="1.1.1.2",
                                        ranges=[
                                            dict(
                                                prefix="2001:db4::/32", cost=11
                                            ),
                                            dict(
                                                prefix="2001:db5::/32",
                                                not_advertise=True,
                                            ),
                                            dict(
                                                prefix="2001:db7::/32",
                                                not_advertise=True,
                                                cost=18,
                                            ),
                                        ],
                                    ),
                                ],
                            ),
                        )
                    ]
                ),
                state="merged",
            ),
            ignore_provider_arg,
        )
        commands = [
            "router ospfv3 100",
            "address-family ipv6 unicast",
            "area 1.1.1.1 range 2001:db3::/32 cost 20",
            "area 1.1.1.2 range 2001:db4::/32 cost 11",
            "area 1.1.1.2 range 2001:db5::/32 not-advertise",
            "area 1.1.1.2 range 2001:db7::/32 not-advertise cost 18",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_af_areas_ranges_replaced(self):
        # test replaced for config->processes->af->areas->ranges
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              address-family ipv6 unicast
                area 1.1.1.1 range 2001:db2::/32
                area 1.1.1.1 range 2001:db3::/32 cost 10
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            address_family=dict(
                                afi="ipv6",
                                safi="unicast",
                                areas=[
                                    dict(
                                        area_id="1.1.1.2",
                                        ranges=[
                                            dict(
                                                prefix="2001:db4::/32", cost=11
                                            ),
                                            dict(
                                                prefix="2001:db5::/32",
                                                not_advertise=True,
                                            ),
                                            dict(
                                                prefix="2001:db7::/32",
                                                not_advertise=True,
                                                cost=18,
                                            ),
                                        ],
                                    )
                                ],
                            ),
                        )
                    ]
                ),
                state="replaced",
            ),
            ignore_provider_arg,
        )
        commands = [
            "router ospfv3 100",
            "address-family ipv6 unicast",
            "no area 1.1.1.1 range 2001:db2::/32",
            "no area 1.1.1.1 range 2001:db3::/32",
            "area 1.1.1.2 range 2001:db4::/32 cost 11",
            "area 1.1.1.2 range 2001:db5::/32 not-advertise",
            "area 1.1.1.2 range 2001:db7::/32 not-advertise cost 18",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_af_areas_default_cost_merged(self):
        # test merged for config->processes->af->areas->default_cost
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              address-family ipv6 unicast
                area 1.1.1.1 default-cost 10
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            address_family=dict(
                                afi="ipv6",
                                safi="unicast",
                                areas=[
                                    dict(area_id="1.1.1.1", default_cost=12),
                                    dict(area_id="1.1.1.2", default_cost=200),
                                ],
                            ),
                        )
                    ]
                ),
                state="merged",
            ),
            ignore_provider_arg,
        )
        commands = [
            "router ospfv3 100",
            "address-family ipv6 unicast",
            "area 1.1.1.1 default-cost 12",
            "area 1.1.1.2 default-cost 200",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_af_areas_default_cost_replaced(self):
        # test merged for config->processes->af->areas->default_cost
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              address-family ipv6 unicast
                area 1.1.1.1 default-cost 10
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            address_family=dict(
                                afi="ipv6",
                                safi="unicast",
                                areas=[
                                    dict(area_id="1.1.1.2", default_cost=200)
                                ],
                            ),
                        )
                    ]
                ),
                state="replaced",
            ),
            ignore_provider_arg,
        )
        commands = [
            "router ospfv3 100",
            "address-family ipv6 unicast",
            "no area 1.1.1.1 default-cost 10",
            "area 1.1.1.2 default-cost 200",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_af_default_information_merged(self):
        # test merged for config->processes->af->default_information
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              address-family ipv6 unicast
                default-information originate
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            address_family=dict(
                                afi="ipv6",
                                safi="unicast",
                                default_information=dict(
                                    originate=dict(
                                        always=True, route_map="test-2"
                                    )
                                ),
                            ),
                        )
                    ]
                ),
                state="merged",
            ),
            ignore_provider_arg,
        )
        commands = [
            "router ospfv3 100",
            "address-family ipv6 unicast",
            "default-information originate always route-map test-2",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_af_default_information_merged_2(self):
        # test merged for config->processes->af->default_information->set
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              address-family ipv6 unicast
                default-information originate always route-map test-2
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            address_family=dict(
                                afi="ipv6",
                                safi="unicast",
                                default_information=dict(
                                    originate=dict(set=False)
                                ),
                            ),
                        )
                    ]
                ),
                state="merged",
            ),
            ignore_provider_arg,
        )
        commands = [
            "router ospfv3 100",
            "address-family ipv6 unicast",
            "no default-information originate",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_af_default_information_replaced(self):
        # test replaced for config->processes->af->default_information
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              address-family ipv6 unicast
                default-information originate always test-2
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            address_family=dict(
                                afi="ipv6",
                                safi="unicast",
                                default_information=dict(
                                    originate=dict(set=True)
                                ),
                            ),
                        )
                    ]
                ),
                state="replaced",
            ),
            ignore_provider_arg,
        )
        commands = [
            "router ospfv3 100",
            "address-family ipv6 unicast",
            "default-information originate",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_af_distance_merged(self):
        # test merged for config->processes->af->distance
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              address-family ipv6 unicast
                distance 20
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            address_family=dict(
                                afi="ipv6", safi="unicast", distance=35
                            ),
                        )
                    ]
                ),
                state="merged",
            ),
            ignore_provider_arg,
        )
        commands = [
            "router ospfv3 100",
            "address-family ipv6 unicast",
            "distance 35",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_af_distance_replaced(self):
        # test replaced for config->processes->af->distance
        # `distance` is a linear attribute so replaced test
        # can only be removal of this attribute
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              address-family ipv6 unicast
                distance 20
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            address_family=dict(afi="ipv6", safi="unicast"),
                        )
                    ]
                ),
                state="replaced",
            ),
            ignore_provider_arg,
        )
        commands = [
            "router ospfv3 100",
            "address-family ipv6 unicast",
            "no distance 20",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_af_maximum_paths_merged(self):
        # test merged for config->processes->af->maximum_paths
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              address-family ipv6 unicast
                maximum-paths 18
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            address_family=dict(
                                afi="ipv6", safi="unicast", maximum_paths=27
                            ),
                        )
                    ]
                ),
                state="replaced",
            ),
            ignore_provider_arg,
        )
        commands = [
            "router ospfv3 100",
            "address-family ipv6 unicast",
            "maximum-paths 27",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_af_maximum_paths_replaced(self):
        # test replaced for config->processes->af->maximum_paths
        # `maximum_paths` is a linear attribute so replaced test
        # can only be removal of this attribute
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              address-family ipv6 unicast
                maximum-paths 18
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            address_family=dict(afi="ipv6", safi="unicast"),
                        )
                    ]
                ),
                state="replaced",
            ),
            ignore_provider_arg,
        )
        commands = [
            "router ospfv3 100",
            "address-family ipv6 unicast",
            "no maximum-paths 18",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_af_redistribute_merged(self):
        # test merged for config->processes->af->redistribute
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              address-family ipv6 unicast
                redistribute eigrp 100 route-map test-17
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            address_family=dict(
                                afi="ipv6",
                                safi="unicast",
                                redistribute=[
                                    dict(
                                        protocol="eigrp",
                                        id="100",
                                        route_map="test-1",
                                    ),
                                    dict(
                                        protocol="eigrp",
                                        id="101",
                                        route_map="test-2",
                                    ),
                                    dict(
                                        protocol="bgp",
                                        id="65563",
                                        route_map="test-3",
                                    ),
                                    dict(
                                        protocol="static", route_map="test-4"
                                    ),
                                ],
                            ),
                        )
                    ]
                ),
                state="merged",
            ),
            ignore_provider_arg,
        )
        commands = [
            "router ospfv3 100",
            "address-family ipv6 unicast",
            "redistribute eigrp 100 route-map test-1",
            "redistribute eigrp 101 route-map test-2",
            "redistribute bgp 65563 route-map test-3",
            "redistribute static route-map test-4",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_af_redistribute_replaced(self):
        # test replaced for config->processes->af->redistribute
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              address-family ipv6 unicast
                redistribute eigrp 100 route-map test-1
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            address_family=dict(
                                afi="ipv6",
                                safi="unicast",
                                redistribute=[
                                    dict(
                                        protocol="eigrp",
                                        id="101",
                                        route_map="test-2",
                                    ),
                                    dict(
                                        protocol="bgp",
                                        id="65563",
                                        route_map="test-3",
                                    ),
                                    dict(
                                        protocol="static", route_map="test-4"
                                    ),
                                ],
                            ),
                        )
                    ]
                ),
                state="replaced",
            ),
            ignore_provider_arg,
        )
        commands = [
            "router ospfv3 100",
            "address-family ipv6 unicast",
            "no redistribute eigrp 100 route-map test-1",
            "redistribute eigrp 101 route-map test-2",
            "redistribute bgp 65563 route-map test-3",
            "redistribute static route-map test-4",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_af_summary_address_merged(self):
        # test merged for config->processes->af->summary_address
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              address-family ipv6 unicast
                summary-address 2001:db2::/32
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            address_family=dict(
                                afi="ipv6",
                                safi="unicast",
                                summary_address=[
                                    dict(prefix="2001:db2::/32", tag=19),
                                    dict(
                                        prefix="2001:db3::/32",
                                        not_advertise=True,
                                    ),
                                    dict(prefix="2001:db4::/32"),
                                ],
                            ),
                        )
                    ]
                ),
                state="merged",
            ),
            ignore_provider_arg,
        )
        commands = [
            "router ospfv3 100",
            "address-family ipv6 unicast",
            "summary-address 2001:db2::/32 tag 19",
            "summary-address 2001:db3::/32 not-advertise",
            "summary-address 2001:db4::/32",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_af_summary_address_replaced(self):
        # test replaced for config->processes->af->summary_address
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              address-family ipv6 unicast
                summary-address 2001:db2::/32 tag 19
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            address_family=dict(
                                afi="ipv6",
                                safi="unicast",
                                summary_address=[
                                    dict(
                                        prefix="2001:db3::/32",
                                        not_advertise=True,
                                    )
                                ],
                            ),
                        )
                    ]
                ),
                state="replaced",
            ),
            ignore_provider_arg,
        )
        commands = [
            "router ospfv3 100",
            "address-family ipv6 unicast",
            "no summary-address 2001:db2::/32 tag 19",
            "summary-address 2001:db3::/32 not-advertise",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_af_table_map_merged(self):
        # test merged for config->processes->af->table_map
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              address-family ipv6 unicast
                summary-address 2001:db2::/32 tag 19
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            address_family=dict(
                                afi="ipv6",
                                safi="unicast",
                                table_map=dict(name="test-1", filter=True),
                            ),
                        )
                    ]
                ),
                state="merged",
            ),
            ignore_provider_arg,
        )
        commands = [
            "router ospfv3 100",
            "address-family ipv6 unicast",
            "table-map test-1 filter",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_af_table_map_replaced(self):
        # test replaced for config->processes->af->table_map
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              address-family ipv6 unicast
                table-map test-1 filter
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            address_family=dict(
                                afi="ipv6",
                                safi="unicast",
                                table_map=dict(name="test-2"),
                            ),
                        )
                    ]
                ),
                state="replaced",
            ),
            ignore_provider_arg,
        )
        commands = [
            "router ospfv3 100",
            "address-family ipv6 unicast",
            "table-map test-2",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_af_timers_merged(self):
        # test merged for config->processes->af->timers
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              address-family ipv6 unicast
                timers throttle spf 1000 20 2800
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            address_family=dict(
                                afi="ipv6",
                                safi="unicast",
                                timers=dict(
                                    throttle=dict(
                                        spf=dict(
                                            initial_spf_delay=1100,
                                            max_wait_time=2805,
                                        )
                                    )
                                ),
                            ),
                        )
                    ]
                ),
                state="merged",
            ),
            ignore_provider_arg,
        )
        commands = [
            "router ospfv3 100",
            "address-family ipv6 unicast",
            "timers throttle spf 1100 20 2805",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_af_timers_replaced(self):
        # test replaced for config->processes->af->timers
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              address-family ipv6 unicast
                timers throttle spf 1000 20 2800
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            address_family=dict(afi="ipv6", safi="unicast"),
                        )
                    ]
                ),
                state="replaced",
            ),
            ignore_provider_arg,
        )
        commands = [
            "router ospfv3 100",
            "address-family ipv6 unicast",
            "no timers throttle spf 1000 20 2800",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_areas_nssa_merged(self):
        # test merged for config->processes->areas->nssa
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              area 1.1.1.1 nssa no-redistribution default-information-originate
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            areas=[
                                dict(
                                    area_id="1.1.1.1",
                                    nssa=dict(no_summary=True),
                                ),
                                dict(area_id="1.1.1.2", nssa=dict(set=True)),
                                dict(
                                    area_id="1.1.1.3",
                                    nssa=dict(
                                        default_information_originate=True,
                                        no_summary=True,
                                        no_redistribution=True,
                                        route_map="test-1",
                                        translate=dict(
                                            type7=dict(
                                                always=True, supress_fa=True
                                            )
                                        ),
                                    ),
                                ),
                            ],
                        )
                    ]
                ),
                state="merged",
            ),
            ignore_provider_arg,
        )
        commands = [
            "router ospfv3 100",
            "area 1.1.1.1 nssa no-summary no-redistribution default-information-originate",
            "area 1.1.1.2 nssa",
            "area 1.1.1.3 nssa translate type7 always supress-fa",
            "area 1.1.1.3 nssa no-summary no-redistribution default-information-originate route-map test-1",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_areas_nssa_merged_2(self):
        # test merged for config->processes->areas->nssa->set
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              area 1.1.1.1 nssa no-summary
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            areas=[
                                dict(area_id="1.1.1.1", nssa=dict(set=False))
                            ],
                        )
                    ]
                ),
                state="merged",
            ),
            ignore_provider_arg,
        )
        commands = ["router ospfv3 100", "no area 1.1.1.1 nssa"]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_areas_nssa_replaced(self):
        # test replaced for config->processes->areas->nssa
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              area 1.1.1.1 nssa no-summary no-redistribution default-information-originate
              area 1.1.1.3 nssa translate type7 always supress-fa
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            areas=[
                                dict(
                                    area_id="1.1.1.3",
                                    nssa=dict(
                                        default_information_originate=True,
                                        no_summary=True,
                                        no_redistribution=True,
                                        route_map="test-1",
                                    ),
                                )
                            ],
                        )
                    ]
                ),
                state="replaced",
            ),
            ignore_provider_arg,
        )
        commands = [
            "router ospfv3 100",
            "no area 1.1.1.1 nssa",
            "no area 1.1.1.3 nssa translate type7 always supress-fa",
            "area 1.1.1.3 nssa no-summary no-redistribution default-information-originate route-map test-1",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_areas_stub_merged(self):
        # test merged for config->processes->areas->stub
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              area 1.1.1.3 stub
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            areas=[
                                dict(
                                    area_id="1.1.1.3",
                                    stub=dict(no_summary=True),
                                )
                            ],
                        )
                    ]
                ),
                state="merged",
            ),
            ignore_provider_arg,
        )
        commands = ["router ospfv3 100", "area 1.1.1.3 stub no-summary"]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_areas_stub_merged_2(self):
        # test merged for config->processes->areas->stub->set
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              area 1.1.1.3 stub
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            areas=[
                                dict(area_id="1.1.1.3", stub=dict(set=False)),
                                dict(
                                    area_id="1.1.1.4",
                                    stub=dict(no_summary=True),
                                ),
                            ],
                        )
                    ]
                ),
                state="merged",
            ),
            ignore_provider_arg,
        )
        commands = [
            "router ospfv3 100",
            "no area 1.1.1.3 stub",
            "area 1.1.1.4 stub no-summary",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_areas_stub_replaced(self):
        # test replaced for config->processes->areas->stub
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              area 1.1.1.3 stub no-summary
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(process_id="100", areas=[dict(area_id="1.1.1.3")])
                    ]
                ),
                state="replaced",
            ),
            ignore_provider_arg,
        )
        commands = ["router ospfv3 100", "no area 1.1.1.3 stub"]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_autocost_flush_route_isolate_merged(self):
        # test merged for config->processes->autocost,flush_routes, isolate
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              auto-cost reference-bandwidth 300 Mbps
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            auto_cost=dict(
                                reference_bandwidth=100, unit="Gbps"
                            ),
                            flush_routes=True,
                            isolate=True,
                        )
                    ]
                ),
                state="merged",
            ),
            ignore_provider_arg,
        )
        commands = [
            "router ospfv3 100",
            "auto-cost reference-bandwidth 100 Gbps",
            "flush-routes",
            "isolate",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_autocost_flush_route_isolate_replaced(self):
        # test merged for config->processes->autocost,flush_routes, isolate
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              auto-cost reference-bandwidth 300 Mbps
              flush-routes
            """
        )
        set_module_args(
            dict(
                config=dict(processes=[dict(process_id="100", isolate=True)]),
                state="replaced",
            ),
            ignore_provider_arg,
        )
        commands = [
            "router ospfv3 100",
            "no auto-cost reference-bandwidth 300 Mbps",
            "no flush-routes",
            "isolate",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_log_adjacency_changes_name_lookup_passive_interface_merged(
        self
    ):
        # test merged for config->processes->log_adjacency_changes, name_lookup, passive_interface
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              log-adjacency-changes
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            log_adjacency_changes=dict(detail=True),
                            name_lookup=True,
                            passive_interface=dict(default=True),
                        ),
                        dict(
                            process_id="102",
                            log_adjacency_changes=dict(log=True),
                        ),
                    ]
                ),
                state="merged",
            ),
            ignore_provider_arg,
        )
        commands = [
            "router ospfv3 100",
            "log-adjacency-changes detail",
            "name-lookup",
            "passive-interface default",
            "router ospfv3 102",
            "log-adjacency-changes",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_log_adjacency_changes_name_lookup_passive_interface_replaced(
        self
    ):
        # test replaced for config->processes->log_adjacency_changes, name_lookup, passive_interface
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              log-adjacency-changes detail
              name-lookup
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            passive_interface=dict(default=True),
                        )
                    ]
                ),
                state="replaced",
            ),
            ignore_provider_arg,
        )
        commands = [
            "router ospfv3 100",
            "no log-adjacency-changes detail",
            "no name-lookup",
            "passive-interface default",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_max_lsa_router_id_merged(self):
        # test merged for config->processes->max_lsa, router_id
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              router-id 192.168.1.100
              max-lsa 4200 85 ignore-count 10 reset-time 120
            router ospfv3 102
              max-lsa 4200 85 ignore-time 120 ignore-count 12 reset-time 300
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            router_id="192.168.1.100",
                            max_lsa=dict(
                                max_non_self_generated_lsa=4200,
                                threshold=85,
                                ignore_count=100,
                                reset_time=138,
                            ),
                        ),
                        dict(
                            process_id="102",
                            router_id="192.168.2.102",
                            max_lsa=dict(
                                max_non_self_generated_lsa=4200,
                                threshold=85,
                                ignore_time=200,
                                ignore_count=20,
                                reset_time=120,
                            ),
                        ),
                        dict(
                            process_id="103",
                            max_lsa=dict(
                                max_non_self_generated_lsa=4200,
                                warning_only=True,
                            ),
                        ),
                    ]
                ),
                state="merged",
            ),
            ignore_provider_arg,
        )
        commands = [
            "router ospfv3 100",
            "max-lsa 4200 85 ignore-count 100 reset-time 138",
            "router ospfv3 102",
            "router-id 192.168.2.102",
            "max-lsa 4200 85 ignore-time 200 ignore-count 20 reset-time 120",
            "router ospfv3 103",
            "max-lsa 4200 warning-only",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_max_lsa_router_id_replaced(self):
        # test replaced for config->processes->max_lsa, router_id
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              router-id 192.168.1.100
              max-lsa 4200 85 ignore-count 10 reset-time 120
            router ospfv3 102
              max-lsa 4200 85 ignore-time 120 ignore-count 12 reset-time 300
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            max_lsa=dict(
                                max_non_self_generated_lsa=4200,
                                threshold=85,
                                warning_only=True,
                            ),
                        )
                    ]
                ),
                state="replaced",
            ),
            ignore_provider_arg,
        )
        commands = [
            "router ospfv3 100",
            "no router-id 192.168.1.100",
            "max-lsa 4200 85 warning-only",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_graceful_restart_merged(self):
        # test merged for config->processes->graceful_restart
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              no graceful-restart
            router ospfv3 102
              no graceful-restart planned-only
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            graceful_restart=dict(
                                grace_period=50, helper_disable=True
                            ),
                        ),
                        dict(
                            process_id="102",
                            graceful_restart=dict(planned_only=True),
                        ),
                    ]
                ),
                state="merged",
            ),
            ignore_provider_arg,
        )
        commands = [
            "router ospfv3 100",
            "graceful-restart grace-period 50",
            "graceful-restart helper-disable",
            "router ospfv3 102",
            "graceful-restart planned-only",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_graceful_restart_replaced(self):
        # test replaced for config->processes->graceful_restart
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              graceful-restart grace-period 50
              graceful-restart helper-disable
            router ospfv3 102
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            graceful_restart=dict(grace_period=10),
                        ),
                        dict(
                            process_id="102",
                            graceful_restart=dict(helper_disable=True),
                        ),
                    ]
                ),
                state="replaced",
            ),
            ignore_provider_arg,
        )
        commands = [
            "router ospfv3 100",
            "graceful-restart grace-period 10",
            "no graceful-restart helper-disable",
            "router ospfv3 102",
            "graceful-restart helper-disable",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_max_metric_merged(self):
        # test merged for config->processes->max_metric
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              max-metric router-lsa external-lsa 1900
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            max_metric=dict(
                                router_lsa=dict(
                                    external_lsa=dict(max_metric_value=2000),
                                    stub_prefix_lsa=True,
                                    on_startup=dict(set=True),
                                )
                            ),
                        ),
                        dict(
                            process_id="102",
                            max_metric=dict(
                                router_lsa=dict(
                                    inter_area_prefix_lsa=dict(
                                        max_metric_value=1800
                                    )
                                )
                            ),
                        ),
                        dict(
                            process_id="103",
                            max_metric=dict(
                                router_lsa=dict(
                                    on_startup=dict(
                                        wait_period=1200,
                                        wait_for_bgp_asn=65563,
                                    ),
                                    inter_area_prefix_lsa=dict(set=True),
                                )
                            ),
                        ),
                    ]
                ),
                state="merged",
            ),
            ignore_provider_arg,
        )
        commands = [
            "router ospfv3 100",
            "max-metric router-lsa external-lsa 2000 stub-prefix-lsa on-startup",
            "router ospfv3 102",
            "max-metric router-lsa inter-area-prefix-lsa 1800",
            "router ospfv3 103",
            "max-metric router-lsa on-startup 1200 wait-for bgp 65563 inter-area-prefix-lsa",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_max_metric_merged_2(self):
        # test merged for config->processes->max_metric->set
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              max-metric router-lsa inter-area-prefix-lsa 1800
            router ospfv3 103
              max-metric router-lsa on-startup 1200 wait-for bgp 65563 inter-area-prefix-lsa
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            max_metric=dict(router_lsa=dict(set=False)),
                        )
                    ]
                ),
                state="merged",
            ),
            ignore_provider_arg,
        )
        commands = ["router ospfv3 100", "no max-metric router-lsa"]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_max_metric_replaced(self):
        # test replaced for config->processes->max_metric
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
            router ospfv3 102
              max-metric router-lsa inter-area-prefix-lsa 1800
            router ospfv3 103
              max-metric router-lsa on-startup 1200 wait-for bgp 65563 inter-area-prefix-lsa
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            max_metric=dict(
                                router_lsa=dict(
                                    external_lsa=dict(max_metric_value=2000),
                                    stub_prefix_lsa=True,
                                    on_startup=dict(set=True),
                                )
                            ),
                        ),
                        dict(process_id="102"),
                        dict(process_id="103"),
                    ]
                ),
                state="replaced",
            ),
            ignore_provider_arg,
        )
        commands = [
            "router ospfv3 100",
            "max-metric router-lsa external-lsa 2000 stub-prefix-lsa on-startup",
            "router ospfv3 102",
            "no max-metric router-lsa",
            "router ospfv3 103",
            "no max-metric router-lsa",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_timers_shutdown_merged(self):
        # test merged for config->processes->timers, shutdown
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              timers lsa-group-pacing 190
              shutdown
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            timers=dict(
                                lsa_arrival=1200,
                                lsa_group_pacing=210,
                                throttle=dict(
                                    lsa=dict(
                                        start_interval=100,
                                        hold_interval=70,
                                        max_interval=1500,
                                    )
                                ),
                            ),
                            shutdown=False,
                        )
                    ]
                ),
                state="merged",
            ),
            ignore_provider_arg,
        )
        commands = [
            "router ospfv3 100",
            "timers lsa-arrival 1200",
            "timers lsa-group-pacing 210",
            "timers throttle lsa 100 70 1500",
            "no shutdown",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_timers_shutdown_replaced(self):
        # test replaced for config->processes->timers, shutdown
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              timers lsa-arrival 800
              timers lsa-group-pacing 210
              timers throttle lsa 100 70 1500
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            timers=dict(lsa_arrival=1200),
                            shutdown=True,
                        )
                    ]
                ),
                state="replaced",
            ),
            ignore_provider_arg,
        )
        commands = [
            "router ospfv3 100",
            "timers lsa-arrival 1200",
            "no timers lsa-group-pacing 210",
            "no timers throttle lsa 100 70 1500",
            "shutdown",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_vrf_merged(self):
        # test merged for config->processes->vrf
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              vrf blue
                area 1.1.1.1 nssa
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            vrfs=[
                                dict(
                                    vrf="blue",
                                    areas=[
                                        dict(
                                            area_id="1.1.1.1",
                                            nssa=dict(no_summary=True),
                                        ),
                                        dict(
                                            area_id="1.1.1.2",
                                            nssa=dict(set=True),
                                        ),
                                    ],
                                ),
                                dict(
                                    vrf="red",
                                    areas=[
                                        dict(
                                            area_id="1.1.1.3",
                                            nssa=dict(
                                                default_information_originate=True,
                                                no_summary=True,
                                                no_redistribution=True,
                                                route_map="test-1",
                                                translate=dict(
                                                    type7=dict(
                                                        always=True,
                                                        supress_fa=True,
                                                    )
                                                ),
                                            ),
                                        )
                                    ],
                                ),
                            ],
                        ),
                        dict(
                            process_id="103",
                            vrfs=[
                                dict(
                                    vrf="red",
                                    max_metric=dict(
                                        router_lsa=dict(
                                            on_startup=dict(
                                                wait_period=1200,
                                                wait_for_bgp_asn=65563,
                                            ),
                                            inter_area_prefix_lsa=dict(
                                                set=True
                                            ),
                                        )
                                    ),
                                )
                            ],
                        ),
                        dict(
                            process_id="104",
                            vrfs=[
                                dict(
                                    vrf="red",
                                    timers=dict(
                                        lsa_arrival=1200,
                                        lsa_group_pacing=210,
                                        throttle=dict(
                                            lsa=dict(
                                                start_interval=100,
                                                hold_interval=70,
                                                max_interval=1500,
                                            )
                                        ),
                                    ),
                                    shutdown=True,
                                )
                            ],
                        ),
                    ]
                ),
                state="merged",
            ),
            ignore_provider_arg,
        )
        commands = [
            "router ospfv3 100",
            "vrf blue",
            "area 1.1.1.1 nssa no-summary",
            "area 1.1.1.2 nssa",
            "vrf red",
            "area 1.1.1.3 nssa translate type7 always supress-fa",
            "area 1.1.1.3 nssa no-summary no-redistribution default-information-originate route-map test-1",
            "router ospfv3 103",
            "vrf red",
            "max-metric router-lsa on-startup 1200 wait-for bgp 65563 inter-area-prefix-lsa",
            "router ospfv3 104",
            "vrf red",
            "timers lsa-arrival 1200",
            "timers lsa-group-pacing 210",
            "timers throttle lsa 100 70 1500",
            "shutdown",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_vrf_replaced(self):
        # test replaced for config->processes->vrf
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              vrf blue
                area 1.1.1.1 nssa no-summary
                area 1.1.1.2 nssa
              vrf red
                area 1.1.1.3 nssa translate type7 always supress-fa
                area 1.1.1.3 nssa no-summary no-redistribution default-information-originate route-map test-1
            router ospfv3 103
              vrf red
                max-metric router-lsa on-startup 1200 wait-for bgp 65563 inter-area-prefix-lsa
            router ospfv3 104
              vrf red
                timers lsa-arrival 1200
                timers lsa-group-pacing 210
                timers throttle lsa 100 70 1500
                shutdown
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            vrfs=[
                                dict(
                                    vrf="blue",
                                    areas=[
                                        dict(
                                            area_id="1.1.1.1",
                                            nssa=dict(
                                                no_summary=True,
                                                translate=dict(
                                                    type7=dict(
                                                        always=True,
                                                        supress_fa=True,
                                                    )
                                                ),
                                            ),
                                        )
                                    ],
                                )
                            ],
                        )
                    ]
                ),
                state="replaced",
            ),
            ignore_provider_arg,
        )
        commands = [
            "router ospfv3 100",
            "vrf blue",
            "area 1.1.1.1 nssa no-summary",
            "no area 1.1.1.2 nssa",
            "area 1.1.1.1 nssa translate type7 always supress-fa",
            "no vrf red",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_top_spec_af_vrf_merged(self):
        # test merged for every nested level
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              address-family ipv6 unicast
                table-map map1 filter
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            address_family=dict(
                                afi="ipv6",
                                safi="unicast",
                                redistribute=[
                                    dict(
                                        protocol="eigrp",
                                        id="100",
                                        route_map="rmap1",
                                    )
                                ],
                            ),
                            vrfs=[
                                dict(vrf="blue", router_id="10.0.0.2"),
                                dict(
                                    vrf="red",
                                    areas=[
                                        dict(
                                            area_id="1.1.1.1",
                                            nssa=dict(set=True),
                                        )
                                    ],
                                ),
                            ],
                        ),
                        dict(
                            process_id="103",
                            vrfs=[dict(vrf="red", shutdown=True)],
                        ),
                    ]
                ),
                state="merged",
            ),
            ignore_provider_arg,
        )
        commands = [
            "router ospfv3 100",
            "address-family ipv6 unicast",
            "redistribute eigrp 100 route-map rmap1",
            "vrf blue",
            "router-id 10.0.0.2",
            "vrf red",
            "area 1.1.1.1 nssa",
            "router ospfv3 103",
            "vrf red",
            "shutdown",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_top_spec_af_vrf_replaced(self):
        # test replaced for every nested level
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              address-family ipv6 unicast
                table-map map1 filter
                redistribute eigrp 100 route-map rmap1
              vrf blue
                router-id 10.0.0.2
              vrf red
                area 1.1.1.1 nssa
            router ospfv3 103
              vrf red
              shutdown
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            address_family=dict(
                                afi="ipv6",
                                safi="unicast",
                                redistribute=[
                                    dict(
                                        protocol="eigrp",
                                        id="100",
                                        route_map="rmap1",
                                    )
                                ],
                            ),
                        )
                    ]
                ),
                state="replaced",
            ),
            ignore_provider_arg,
        )
        commands = [
            "router ospfv3 100",
            "address-family ipv6 unicast",
            "no table-map map1 filter",
            "no vrf blue",
            "no vrf red",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_top_spec_af_vrf_overridden(self):
        # test overridden for every nested level
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              address-family ipv6 unicast
                table-map map1 filter
                redistribute eigrp 100 route-map rmap1
              vrf blue
                router-id 10.0.0.2
              vrf red
                area 1.1.1.1 nssa
            router ospfv3 103
              vrf red
              shutdown
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(
                            process_id="100",
                            address_family=dict(
                                afi="ipv6",
                                safi="unicast",
                                redistribute=[
                                    dict(
                                        protocol="eigrp",
                                        id="100",
                                        route_map="rmap2",
                                    )
                                ],
                            ),
                        )
                    ]
                ),
                state="overridden",
            ),
            ignore_provider_arg,
        )
        commands = [
            "router ospfv3 100",
            "address-family ipv6 unicast",
            "no table-map map1 filter",
            "redistribute eigrp 100 route-map rmap2",
            "no vrf blue",
            "no vrf red",
            "no router ospfv3 103",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_top_spec_af_vrf_deleted(self):
        # test overridden for every nested level
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              address-family ipv6 unicast
                table-map map1 filter
                redistribute eigrp 100 route-map rmap1
              vrf blue
                router-id 10.0.0.2
              vrf red
                area 1.1.1.1 nssa
            router ospfv3 103
              vrf red
              shutdown
            """
        )
        set_module_args(
            dict(
                config=dict(processes=[dict(process_id="100")]),
                state="deleted",
            ),
            ignore_provider_arg,
        )
        commands = ["no router ospfv3 100"]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_top_spec_af_vrf_deleted_all(self):
        # test overridden for every nested level
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              address-family ipv6 unicast
                table-map map1 filter
                redistribute eigrp 100 route-map rmap1
              vrf blue
                router-id 10.0.0.2
              vrf red
                area 1.1.1.1 nssa
            router ospfv3 103
              vrf red
              shutdown
            """
        )
        set_module_args(dict(state="deleted"), ignore_provider_arg)
        commands = ["no router ospfv3 100", "no router ospfv3 103"]
        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))

    def test_nxos_ospfv3_parsed(self):
        # test parsed
        set_module_args(
            dict(
                running_config=dedent(
                    """\
                    router ospfv3 100
                      address-family ipv6 unicast
                        table-map map1 filter
                        redistribute eigrp 100 route-map rmap1
                      vrf blue
                        router-id 10.0.0.2
                      vrf red
                        area 1.1.1.1 nssa
                    router ospfv3 103
                      vrf red
                        shutdown
                    """
                ),
                state="parsed",
            ),
            ignore_provider_arg,
        )
        parsed = {
            "processes": [
                {
                    "process_id": "100",
                    "address_family": {
                        "table_map": {"name": "map1", "filter": True},
                        "redistribute": [
                            {
                                "protocol": "eigrp",
                                "id": "100",
                                "route_map": "rmap1",
                            }
                        ],
                        "afi": "ipv6",
                        "safi": "unicast",
                    },
                    "vrfs": [
                        {"vrf": "blue", "router_id": "10.0.0.2"},
                        {
                            "vrf": "red",
                            "areas": [
                                {"area_id": "1.1.1.1", "nssa": {"set": True}}
                            ],
                        },
                    ],
                },
                {
                    "process_id": "103",
                    "vrfs": [{"vrf": "red", "shutdown": True}],
                },
            ]
        }
        result = self.execute_module(changed=False)
        self.assertEqual(set(result["parsed"]), set(parsed))

    def test_nxos_ospfv3_gathered(self):
        # test gathered
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              address-family ipv6 unicast
                table-map map1 filter
                redistribute eigrp 100 route-map rmap1
              vrf blue
                router-id 10.0.0.2
              vrf red
                area 1.1.1.1 nssa
            router ospfv3 103
              vrf red
              shutdown
            """
        )
        set_module_args(dict(state="gathered"), ignore_provider_arg)
        gathered = {
            "processes": [
                {
                    "process_id": "100",
                    "address_family": {
                        "table_map": {"name": "map1", "filter": True},
                        "redistribute": [
                            {
                                "protocol": "eigrp",
                                "id": "100",
                                "route_map": "rmap1",
                            }
                        ],
                        "afi": "ipv6",
                        "safi": "unicast",
                    },
                    "vrfs": [
                        {"vrf": "blue", "router_id": "10.0.0.2"},
                        {
                            "vrf": "red",
                            "areas": [
                                {"area_id": "1.1.1.1", "nssa": {"set": True}}
                            ],
                        },
                    ],
                },
                {
                    "process_id": "103",
                    "vrfs": [{"vrf": "red", "shutdown": True}],
                },
            ]
        }
        result = self.execute_module(changed=False)
        self.assertEqual(set(result["gathered"]), set(gathered))

    def test_nxos_ospfv3_process_id_word(self):
        self.get_config.return_value = dedent(
            """\
            router ospfv3 100
              router-id 203.0.113.20
            router ospfv3 TEST-1
              router-id 198.51.100.1
            """
        )
        set_module_args(
            dict(
                config=dict(
                    processes=[
                        dict(process_id="100", router_id="203.0.113.20"),
                        dict(process_id="TEST-1", router_id="198.51.100.1"),
                        dict(process_id="TEST-2", router_id="198.52.200.1"),
                    ]
                ),
                state="merged",
            ),
            ignore_provider_arg,
        )

        commands = ["router ospfv3 TEST-2", "router-id 198.52.200.1"]

        result = self.execute_module(changed=True)
        self.assertEqual(set(result["commands"]), set(commands))



import os
import yaml

import testinfra.utils.ansible_runner

testinfra_hosts = testinfra.utils.ansible_runner.AnsibleRunner(
    os.environ["MOLECULE_INVENTORY_FILE"]
).get_hosts("all")


def include_vars(host):
    ansible = host.ansible("include_vars",
                           'file="../../defaults/main.yml"',
                           False,
                           False)
    return ansible


def test_mongod_cnf_file(host):
    f = host.file("/etc/mongod.conf")

    assert f.exists
    assert yaml.safe_load(f.content)["security"]["authorization"] == "enabled"


def test_mongod_service(host):
    mongod_service = include_vars(host)["ansible_facts"].get("mongod_service", "mongod")
    s = host.service(mongod_service)

    assert s.is_running
    assert s.is_enabled


def test_mongod_port(host):
    port = include_vars(host)["ansible_facts"].get("mongod_port", 27017)
    s = host.socket("tcp://0.0.0.0:{0}".format(port))
    assert s.is_listening


def test_mongo_shell_connectivity(host):
    """
    Tests that we can connect to mongos via the shell annd run a cmd
    """
    facts = include_vars(host)["ansible_facts"]
    port = facts.get("mongod_port", 27017)
    user = facts.get("mongod_admin_user", "admin")
    pwd = facts.get("mongod_default_admin_pwd", "admin")

    cmd = host.run(
        "mongo admin --username {user} --password {pwd} --port {port} --eval 'db.runCommand({{listDatabases: 1}})'".format(
            user=user, pwd=pwd, port=port
        )
    )

    assert cmd.rc == 0
    assert "admin" in cmd.stdout



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2017, Sertac Ozercan <seozerca@microsoft.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = '''
---
module: azure_rm_virtualmachinescaleset_info


short_description: Get Virtual Machine Scale Set facts

description:
    - Get facts for a virtual machine scale set.
    - Note that this module was called M(community.azure.azure_rm_virtualmachine_scaleset_facts) before Ansible 2.8. The usage did not change.

options:
    name:
        description:
            - Limit results to a specific virtual machine scale set.
    resource_group:
        description:
            - The resource group to search for the desired virtual machine scale set.
    tags:
        description:
            - List of tags to be matched.
    format:
        description:
            - Format of the data returned.
            - If C(raw) is selected information will be returned in raw format from Azure Python SDK.
            - If C(curated) is selected the structure will be identical to input parameters of M(community.azure.azure_rm_virtualmachinescaleset) module.
            - In Ansible 2.5 and lower facts are always returned in raw format.
            - Please note that this option will be deprecated in 2.10 when curated format will become the only supported format.
        default: 'raw'
        choices:
            - 'curated'
            - 'raw'

extends_documentation_fragment:
- azure.azcollection.azure


author:
    - Sertac Ozercan (@sozercan)
'''

EXAMPLES = '''
    - name: Get facts for a virtual machine scale set
      community.azure.azure_rm_virtualmachinescaleset_info:
        resource_group: myResourceGroup
        name: testvmss001
        format: curated

    - name: Get facts for all virtual networks
      community.azure.azure_rm_virtualmachinescaleset_info:
        resource_group: myResourceGroup

    - name: Get facts by tags
      community.azure.azure_rm_virtualmachinescaleset_info:
        resource_group: myResourceGroup
        tags:
          - testing
'''

RETURN = '''
vmss:
    description:
        - List of virtual machine scale sets.
    returned: always
    type: complex
    contains:
        id:
            description:
                - Resource ID.
            returned: always
            type: str
            sample: /subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/myResourceGroup/providers/Microsoft.Compute/scalesets/myscaleset
        admin_username:
            description:
                - Admin username used to access the host after it is created.
            returned: always
            type: str
            sample: adminuser
        capacity:
            description:
                - Capacity of VMSS.
            returned: always
            type: int
            sample: 2
        data_disks:
            description:
                - List of attached data disks.
            returned: always
            type: complex
            contains:
                caching:
                    description:
                        - Type of data disk caching.
                    returned: always
                    type: str
                    sample: ReadOnly
                disk_size_gb:
                    description:
                        - The initial disk size in GB for blank data disks.
                    returned: always
                    type: int
                    sample: 64
                lun:
                    description:
                        - The logical unit number for data disk.
                    returned: always
                    type: int
                    sample: 0
                managed_disk_type:
                    description:
                        - Managed data disk type.
                    returned: always
                    type: str
                    sample: Standard_LRS
        image:
            description:
                - Image specification.
            returned: always
            type: complex
            contains:
                offer:
                    description:
                        - The offer of the platform image or marketplace image used to create the virtual machine.
                    returned: always
                    type: str
                    sample: RHEL
                publisher:
                    description:
                        - Publisher name.
                    returned: always
                    type: str
                    sample: RedHat
                sku:
                    description:
                        - SKU name.
                    returned: always
                    type: str
                    sample: 7-RAW
                version:
                    description:
                        - Image version.
                    returned: always
                    type: str
                    sample: 7.5.2018050901
        load_balancer:
            description:
                - Load balancer name.
            returned: always
            type: str
            sample: testlb
        location:
            description:
                - Resource location.
            type: str
            returned: always
            sample: japaneast
        managed_disk_type:
            description:
                - Managed data disk type.
            type: str
            returned: always
            sample: Standard_LRS
        name:
            description:
                - Resource name.
            returned: always
            type: str
            sample: myvmss
        os_disk_caching:
            description:
                - Type of OS disk caching.
            type: str
            returned: always
            sample: ReadOnly
        os_type:
            description:
                - Base type of operating system.
            type: str
            returned: always
            sample: Linux
        overprovision:
            description:
                - Specifies whether the Virtual Machine Scale Set should be overprovisioned.
            type: bool
            sample: true
        resource_group:
            description:
                - Resource group.
            type: str
            returned: always
            sample: myResourceGroup
        ssh_password_enabled:
            description:
                - Is SSH password authentication enabled. Valid only for Linux.
            type: bool
            returned: always
            sample: true
        subnet_name:
            description:
                - Subnet name.
            type: str
            returned: always
            sample: testsubnet
        tier:
            description:
                - SKU Tier.
            type: str
            returned: always
            sample: Basic
        upgrade_policy:
            description:
                - Upgrade policy.
            type: str
            returned: always
            sample: Manual
        virtual_network_name:
            description:
                - Associated virtual network name.
            type: str
            returned: always
            sample: testvn
        vm_size:
            description:
                - Virtual machine size.
            type: str
            returned: always
            sample: Standard_D4
        tags:
            description:
                - Tags assigned to the resource. Dictionary of string:string pairs.
            returned: always
            type: dict
            sample: { "tag1": "abc" }
'''  # NOQA

from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common import AzureRMModuleBase
import re

try:
    from msrestazure.azure_exceptions import CloudError
except Exception:
    # handled in azure_rm_common
    pass

AZURE_OBJECT_CLASS = 'VirtualMachineScaleSet'

AZURE_ENUM_MODULES = ['azure.mgmt.compute.models']


class AzureRMVirtualMachineScaleSetInfo(AzureRMModuleBase):
    """Utility class to get virtual machine scale set facts"""

    def __init__(self):

        self.module_args = dict(
            name=dict(type='str'),
            resource_group=dict(type='str'),
            tags=dict(type='list'),
            format=dict(
                type='str',
                choices=['curated',
                         'raw'],
                default='raw'
            )
        )

        self.results = dict(
            changed=False,
        )

        self.name = None
        self.resource_group = None
        self.format = None
        self.tags = None

        super(AzureRMVirtualMachineScaleSetInfo, self).__init__(
            derived_arg_spec=self.module_args,
            supports_tags=False,
            facts_module=True
        )

    def exec_module(self, **kwargs):
        is_old_facts = self.module._name == 'azure_rm_virtualmachinescaleset_facts'
        if is_old_facts:
            self.module.deprecate("The 'azure_rm_virtualmachinescaleset_facts' module has been renamed to 'azure_rm_virtualmachinescaleset_info'",
                                  version='2.13')

        for key in self.module_args:
            setattr(self, key, kwargs[key])

        if self.name and not self.resource_group:
            self.fail("Parameter error: resource group required when filtering by name.")

        if self.name:
            result = self.get_item()
        else:
            result = self.list_items()

        if self.format == 'curated':
            for index in range(len(result)):
                vmss = result[index]
                subnet_name = None
                load_balancer_name = None
                virtual_network_name = None
                ssh_password_enabled = False

                try:
                    subnet_id = (vmss['properties']['virtualMachineProfile']['networkProfile']['networkInterfaceConfigurations'][0]
                                 ['properties']['ipConfigurations'][0]['properties']['subnet']['id'])
                    subnet_name = re.sub('.*subnets\\/', '', subnet_id)
                except Exception:
                    self.log('Could not extract subnet name')

                try:
                    backend_address_pool_id = (vmss['properties']['virtualMachineProfile']['networkProfile']['networkInterfaceConfigurations'][0]
                                               ['properties']['ipConfigurations'][0]['properties']['loadBalancerBackendAddressPools'][0]['id'])
                    load_balancer_name = re.sub('\\/backendAddressPools.*', '', re.sub('.*loadBalancers\\/', '', backend_address_pool_id))
                    virtual_network_name = re.sub('.*virtualNetworks\\/', '', re.sub('\\/subnets.*', '', subnet_id))
                except Exception:
                    self.log('Could not extract load balancer / virtual network name')

                try:
                    ssh_password_enabled = (not vmss['properties']['virtualMachineProfile']['osProfile']
                                                    ['linuxConfiguration']['disablePasswordAuthentication'])
                except Exception:
                    self.log('Could not extract SSH password enabled')

                data_disks = vmss['properties']['virtualMachineProfile']['storageProfile'].get('dataDisks', [])

                for disk_index in range(len(data_disks)):
                    old_disk = data_disks[disk_index]
                    new_disk = {
                        'lun': old_disk['lun'],
                        'disk_size_gb': old_disk['diskSizeGB'],
                        'managed_disk_type': old_disk['managedDisk']['storageAccountType'],
                        'caching': old_disk['caching']
                    }
                    data_disks[disk_index] = new_disk

                updated = {
                    'id': vmss['id'],
                    'resource_group': self.resource_group,
                    'name': vmss['name'],
                    'state': 'present',
                    'location': vmss['location'],
                    'vm_size': vmss['sku']['name'],
                    'capacity': vmss['sku']['capacity'],
                    'tier': vmss['sku']['tier'],
                    'upgrade_policy': vmss['properties']['upgradePolicy']['mode'],
                    'admin_username': vmss['properties']['virtualMachineProfile']['osProfile']['adminUsername'],
                    'admin_password': vmss['properties']['virtualMachineProfile']['osProfile'].get('adminPassword'),
                    'ssh_password_enabled': ssh_password_enabled,
                    'image': vmss['properties']['virtualMachineProfile']['storageProfile']['imageReference'],
                    'os_disk_caching': vmss['properties']['virtualMachineProfile']['storageProfile']['osDisk']['caching'],
                    'os_type': 'Linux' if (vmss['properties']['virtualMachineProfile']['osProfile'].get('linuxConfiguration') is not None) else 'Windows',
                    'overprovision': vmss['properties']['overprovision'],
                    'managed_disk_type': vmss['properties']['virtualMachineProfile']['storageProfile']['osDisk']['managedDisk']['storageAccountType'],
                    'data_disks': data_disks,
                    'virtual_network_name': virtual_network_name,
                    'subnet_name': subnet_name,
                    'load_balancer': load_balancer_name,
                    'tags': vmss.get('tags')
                }

                result[index] = updated

        if is_old_facts:
            self.results['ansible_facts'] = {
                'azure_vmss': result
            }
            if self.format == 'curated':
                # proper result format we want to support in the future
                # dropping 'ansible_facts' and shorter name 'vmss'
                self.results['vmss'] = result
        else:
            self.results['vmss'] = result

        return self.results

    def get_item(self):
        """Get a single virtual machine scale set"""

        self.log('Get properties for {0}'.format(self.name))

        item = None
        results = []

        try:
            item = self.compute_client.virtual_machine_scale_sets.get(self.resource_group, self.name)
        except CloudError:
            pass

        if item and self.has_tags(item.tags, self.tags):
            results = [self.serialize_obj(item, AZURE_OBJECT_CLASS, enum_modules=AZURE_ENUM_MODULES)]

        return results

    def list_items(self):
        """Get all virtual machine scale sets"""

        self.log('List all virtual machine scale sets')

        try:
            response = self.compute_client.virtual_machine_scale_sets.list(self.resource_group)
        except CloudError as exc:
            self.fail('Failed to list all items - {0}'.format(str(exc)))

        results = []
        for item in response:
            if self.has_tags(item.tags, self.tags):
                results.append(self.serialize_obj(item, AZURE_OBJECT_CLASS, enum_modules=AZURE_ENUM_MODULES))

        return results


def main():
    """Main module execution code path"""

    AzureRMVirtualMachineScaleSetInfo()


if __name__ == '__main__':
    main()



#!/usr/bin/python
# coding: utf-8
#
# Copyright 2017 Red Hat | Ansible, Alex Grönholm <alex.gronholm@nextday.fi>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
module: docker_volume
short_description: Manage Docker volumes
description:
  - Create/remove Docker volumes.
  - Performs largely the same function as the "docker volume" CLI subcommand.
options:
  volume_name:
    description:
      - Name of the volume to operate on.
    type: str
    required: yes
    aliases:
      - name

  driver:
    description:
      - Specify the type of volume. Docker provides the C(local) driver, but 3rd party drivers can also be used.
    type: str
    default: local

  driver_options:
    description:
      - "Dictionary of volume settings. Consult docker docs for valid options and values:
        U(https://docs.docker.com/engine/reference/commandline/volume_create/#driver-specific-options)"
    type: dict

  labels:
    description:
      - Dictionary of label key/values to set for the volume
    type: dict

  force:
    description:
      - With state C(present) causes the volume to be deleted and recreated if the volume already
        exist and the driver, driver options or labels differ. This will cause any data in the existing
        volume to be lost.
      - Deprecated. Will be removed in community.general 2.0.0. Set I(recreate) to C(options-changed) instead
        for the same behavior of setting I(force) to C(yes).
    type: bool

  recreate:
    description:
      - Controls when a volume will be recreated when I(state) is C(present). Please
        note that recreating an existing volume will cause **any data in the existing volume
        to be lost!** The volume will be deleted and a new volume with the same name will be
        created.
      - The value C(always) forces the volume to be always recreated.
      - The value C(never) makes sure the volume will not be recreated.
      - The value C(options-changed) makes sure the volume will be recreated if the volume
        already exist and the driver, driver options or labels differ.
    type: str
    default: never
    choices:
    - always
    - never
    - options-changed

  state:
    description:
      - C(absent) deletes the volume.
      - C(present) creates the volume, if it does not already exist.
    type: str
    default: present
    choices:
      - absent
      - present

extends_documentation_fragment:
- community.general.docker
- community.general.docker.docker_py_1_documentation


author:
  - Alex Grönholm (@agronholm)

requirements:
  - "L(Docker SDK for Python,https://docker-py.readthedocs.io/en/stable/) >= 1.10.0 (use L(docker-py,https://pypi.org/project/docker-py/) for Python 2.6)"
  - "The docker server >= 1.9.0"
'''

EXAMPLES = '''
- name: Create a volume
  community.general.docker_volume:
    name: volume_one

- name: Remove a volume
  community.general.docker_volume:
    name: volume_one
    state: absent

- name: Create a volume with options
  community.general.docker_volume:
    name: volume_two
    driver_options:
      type: btrfs
      device: /dev/sda2
'''

RETURN = '''
volume:
    description:
    - Volume inspection results for the affected volume.
    - Note that facts are part of the registered vars since Ansible 2.8. For compatibility reasons, the facts
      are also accessible directly as C(docker_volume). Note that the returned fact will be removed in community.general 2.0.0.
    returned: success
    type: dict
    sample: {}
'''

import traceback

try:
    from docker.errors import DockerException, APIError
except ImportError:
    # missing Docker SDK for Python handled in ansible.module_utils.docker.common
    pass

from ansible_collections.community.general.plugins.module_utils.docker.common import (
    DockerBaseClass,
    AnsibleDockerClient,
    DifferenceTracker,
    RequestException,
)
from ansible.module_utils.six import iteritems, text_type


class TaskParameters(DockerBaseClass):
    def __init__(self, client):
        super(TaskParameters, self).__init__()
        self.client = client

        self.volume_name = None
        self.driver = None
        self.driver_options = None
        self.labels = None
        self.force = None
        self.recreate = None
        self.debug = None

        for key, value in iteritems(client.module.params):
            setattr(self, key, value)

        if self.force is not None:
            if self.recreate != 'never':
                client.fail('Cannot use the deprecated "force" '
                            'option when "recreate" is set. Please stop '
                            'using the force option.')
            client.module.warn('The "force" option of docker_volume has been deprecated '
                               'in Ansible 2.8. Please use the "recreate" '
                               'option, which provides the same functionality as "force".')
            self.recreate = 'options-changed' if self.force else 'never'


class DockerVolumeManager(object):

    def __init__(self, client):
        self.client = client
        self.parameters = TaskParameters(client)
        self.check_mode = self.client.check_mode
        self.results = {
            u'changed': False,
            u'actions': []
        }
        self.diff = self.client.module._diff
        self.diff_tracker = DifferenceTracker()
        self.diff_result = dict()

        self.existing_volume = self.get_existing_volume()

        state = self.parameters.state
        if state == 'present':
            self.present()
        elif state == 'absent':
            self.absent()

        if self.diff or self.check_mode or self.parameters.debug:
            if self.diff:
                self.diff_result['before'], self.diff_result['after'] = self.diff_tracker.get_before_after()
            self.results['diff'] = self.diff_result

    def get_existing_volume(self):
        try:
            volumes = self.client.volumes()
        except APIError as e:
            self.client.fail(text_type(e))

        if volumes[u'Volumes'] is None:
            return None

        for volume in volumes[u'Volumes']:
            if volume['Name'] == self.parameters.volume_name:
                return volume

        return None

    def has_different_config(self):
        """
        Return the list of differences between the current parameters and the existing volume.

        :return: list of options that differ
        """
        differences = DifferenceTracker()
        if self.parameters.driver and self.parameters.driver != self.existing_volume['Driver']:
            differences.add('driver', parameter=self.parameters.driver, active=self.existing_volume['Driver'])
        if self.parameters.driver_options:
            if not self.existing_volume.get('Options'):
                differences.add('driver_options',
                                parameter=self.parameters.driver_options,
                                active=self.existing_volume.get('Options'))
            else:
                for key, value in iteritems(self.parameters.driver_options):
                    if (not self.existing_volume['Options'].get(key) or
                            value != self.existing_volume['Options'][key]):
                        differences.add('driver_options.%s' % key,
                                        parameter=value,
                                        active=self.existing_volume['Options'].get(key))
        if self.parameters.labels:
            existing_labels = self.existing_volume.get('Labels', {})
            for label in self.parameters.labels:
                if existing_labels.get(label) != self.parameters.labels.get(label):
                    differences.add('labels.%s' % label,
                                    parameter=self.parameters.labels.get(label),
                                    active=existing_labels.get(label))

        return differences

    def create_volume(self):
        if not self.existing_volume:
            if not self.check_mode:
                try:
                    params = dict(
                        driver=self.parameters.driver,
                        driver_opts=self.parameters.driver_options,
                    )

                    if self.parameters.labels is not None:
                        params['labels'] = self.parameters.labels

                    resp = self.client.create_volume(self.parameters.volume_name, **params)
                    self.existing_volume = self.client.inspect_volume(resp['Name'])
                except APIError as e:
                    self.client.fail(text_type(e))

            self.results['actions'].append("Created volume %s with driver %s" % (self.parameters.volume_name, self.parameters.driver))
            self.results['changed'] = True

    def remove_volume(self):
        if self.existing_volume:
            if not self.check_mode:
                try:
                    self.client.remove_volume(self.parameters.volume_name)
                except APIError as e:
                    self.client.fail(text_type(e))

            self.results['actions'].append("Removed volume %s" % self.parameters.volume_name)
            self.results['changed'] = True

    def present(self):
        differences = DifferenceTracker()
        if self.existing_volume:
            differences = self.has_different_config()

        self.diff_tracker.add('exists', parameter=True, active=self.existing_volume is not None)
        if (not differences.empty and self.parameters.recreate == 'options-changed') or self.parameters.recreate == 'always':
            self.remove_volume()
            self.existing_volume = None

        self.create_volume()

        if self.diff or self.check_mode or self.parameters.debug:
            self.diff_result['differences'] = differences.get_legacy_docker_diffs()
            self.diff_tracker.merge(differences)

        if not self.check_mode and not self.parameters.debug:
            self.results.pop('actions')

        volume_facts = self.get_existing_volume()
        self.results['ansible_facts'] = {u'docker_volume': volume_facts}
        self.results['volume'] = volume_facts

    def absent(self):
        self.diff_tracker.add('exists', parameter=False, active=self.existing_volume is not None)
        self.remove_volume()


def main():
    argument_spec = dict(
        volume_name=dict(type='str', required=True, aliases=['name']),
        state=dict(type='str', default='present', choices=['present', 'absent']),
        driver=dict(type='str', default='local'),
        driver_options=dict(type='dict', default={}),
        labels=dict(type='dict'),
        force=dict(type='bool', removed_in_version='2.0.0', removed_from_collection='community.general'),  # was Ansible 2.12
        recreate=dict(type='str', default='never', choices=['always', 'never', 'options-changed']),
        debug=dict(type='bool', default=False)
    )

    option_minimal_versions = dict(
        labels=dict(docker_py_version='1.10.0', docker_api_version='1.23'),
    )

    client = AnsibleDockerClient(
        argument_spec=argument_spec,
        supports_check_mode=True,
        min_docker_version='1.10.0',
        min_docker_api_version='1.21',
        # "The docker server >= 1.9.0"
        option_minimal_versions=option_minimal_versions,
    )

    try:
        cm = DockerVolumeManager(client)
        client.module.exit_json(**cm.results)
    except DockerException as e:
        client.fail('An unexpected docker error occurred: {0}'.format(e), exception=traceback.format_exc())
    except RequestException as e:
        client.fail('An unexpected requests error occurred when docker-py tried to talk to the docker daemon: {0}'.format(e), exception=traceback.format_exc())


if __name__ == '__main__':
    main()
